#//----------------------------------------------------------------------
#//   Copyright 2011 Cypress Semiconductor
#//   Copyright 2010-2011 Mentor Graphics Corporation
#//   Copyright 2014 NVIDIA Corporation
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
"""
Title: UVM Configuration Database

Topic: Intro

Avoid using `UVMConfigDb` if possible. Instead, group the configuration values
into configuration objects (object per agent), and use hierarchical
configuration objects for passing the configuration from test-level to
env-level, and from there to agents and other sub-components.

The `UVMConfigDb` class provides a convenience interface
on top of the `UVMResourceDb` to simplify the basic interface
that is used for configuring `UVMComponent` instances.

If the run-time `+UVM_CONFIG_DB_TRACE` command line option is specified,
all configuration DB accesses (read and write) are displayed.

NOTE: `UVMConfigDb` is not required for passing interfaces to the verification
components. In uvm-python, you can use `run_test(..., dut=dut)` inside your `@cocotb.test()`
function instead.
"""

from cocotb.triggers import Event

from .uvm_resource import UVMResourceBase, UVMResource, UVMResourcePool, PRI_HIGH
from .uvm_resource_db import UVMResourceDb
from .uvm_pool import UVMPool
from .sv import uvm_re_match, uvm_glob_to_re
from .uvm_debug import uvm_debug


class m_uvm_waiter:

    def __init__(self, inst_name, field_name):
        """
        Args:
            inst_name:
            field_name:
        """
        self.inst_name = inst_name
        self.field_name = field_name
        self.trigger = Event()


class UVMConfigDb(UVMResourceDb):
    """
     All of the functions in UVMConfigDb are static, so they
     must be called using the classname.  For example::

     UVMConfigDb.set(self, "*", "A");

     The parameter value "int" identifies the configuration type as
     an int property.

     The `UVMConfigDb.set` and `UVMConfigDb.get` methods provide the same API and
     semantics as the set/get_config_* functions in `UVMComponent`.
    """

    #  // Internal lookup of config settings so they can be reused
    #  // The context has a pool that is keyed by the inst/field name.
    #  static uvm_pool#(string,uvm_resource#(T)) m_rsc[UVMComponent];
    m_rsc = {}

    #  // Internal waiter list for wait_modified
    #  static local uvm_queue#(m_uvm_waiter) m_waiters[string];
    m_waiters = {}

    @classmethod
    def get(cls, cntxt, inst_name, field_name, value, T=None):
        """
        Get the value for `field_name` in `inst_name`, using component `cntxt` as
        the starting search point. `inst_name` is an explicit instance name
        relative to `cntxt` and may be an empty string if the `cntxt` is the
        instance that the configuration object applies to. `field_name`
        is the specific field in the scope that is being searched for.

        The basic ~get_config_*~ methods from `UVMComponent` are mapped to
        this function as:

        .. code-block:: python

          get_config_int(...) => uvm_config_db#(uvm_bitstream_t)::get(cntxt,...)
          get_config_string(...) => uvm_config_db#(string)::get(cntxt,...)
          get_config_object(...) => uvm_config_db#(uvm_object)::get(cntxt,...)
          @classmethod
          def bit get(cls,  cntxt, inst_name, field_name, value):

        Args:
            cntxt:
            inst_name:
            field_name:
            value:
            T:
        Returns:
        Raises:
        """
        #//TBD: add file/line
        # p = 0
        r = None  # uvm_resource#(T) r, rt;
        # rt = None
        rp = UVMResourcePool.get()
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()

        if cntxt is None:
            cntxt = cs.get_root()
        if inst_name == "":
            inst_name = cntxt.get_full_name()
        elif cntxt.get_full_name() != "":
            inst_name = cntxt.get_full_name() + "." + inst_name

        #rq = rp.lookup_regex_names(inst_name, field_name, uvm_resource#(T)::get_type());
        rq = rp.lookup_regex_names(inst_name, field_name)
        uvm_debug(cls, 'get', 'rq size is ' + str(rq))
        r = UVMResource.get_highest_precedence(rq, T)

        if UVMConfigDbOptions.is_tracing():
            UVMResourceDb.m_show_msg("CFGDB/GET", "Configuration","read", inst_name,
                field_name, cntxt, r)
        if r is None:
            return False

        if hasattr(value, 'append'):
            value.append(r.read(cntxt))
        else:
            raise Exception('value must be a list-like with append')
        return True

    @classmethod
    def set(cls, cntxt, inst_name, field_name, value, T=None):
        """
        Create a new or update an existing configuration setting for
        `field_name` in `inst_name` from `cntxt`.
        The setting is made at `cntxt`, with the full scope of the set
        being {`cntxt`,".",`inst_name`}. If `cntxt` is `null` then `inst_name`
        provides the complete scope information of the setting.
        `field_name` is the target field. Both `inst_name` and `field_name`
        may be glob style or regular expression style expressions.

        If a setting is made at build time, the `cntxt` hierarchy is
        used to determine the setting's precedence in the database.
        Settings from hierarchically higher levels have higher
        precedence. Settings from the same level of hierarchy have
        a last setting wins semantic. A precedence setting of
        <uvm_resource_base::default_precedence>  is used for uvm_top, and
        each hierarchical level below the top is decremented by 1.

        After build time, all settings use the default precedence and thus
        have a last wins semantic. So, if at run time, a low level
        component makes a runtime setting of some field, that setting
        will have precedence over a setting from the test level that was
        made earlier in the simulation.

        The basic ~set_config_*~ methods from `UVMComponent` are mapped to
        this function as:

        .. code-block:: python

          set_config_int(...) => uvm_config_db#(uvm_bitstream_t)::set(cntxt,...)
          set_config_string(...) => uvm_config_db#(string)::set(cntxt,...)
          set_config_object(...) => uvm_config_db#(uvm_object)::set(cntxt,...)

        Args:
            cntxt:
            inst_name:
            field_name:
            value:
            T:
        """
        p = None
        curr_phase = None
        exists = False
        lookup = ""
        pool = None  # uvm_pool#(string,uvm_resource#(T)) pool;
        rstate = ""
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()

        # take care of random stability during allocation
        # process p = process::self();
        # if(p != null)
        # rstate = p.get_randstate();

        top = cs.get_root()
        curr_phase = top.m_current_phase

        if cntxt is None:
            cntxt = top
        if inst_name == "":
            inst_name = cntxt.get_full_name()
        elif cntxt.get_full_name() != "":
            inst_name = cntxt.get_full_name() + "." + inst_name

        if cntxt in UVMConfigDb.m_rsc:
            pass
        else:
            UVMConfigDb.m_rsc[cntxt] = UVMPool()
        pool = UVMConfigDb.m_rsc[cntxt]

        # Insert the token in the middle to prevent cache
        # oddities like i=foobar,f=xyz and i=foo,f=barxyz.
        # Can't just use '.', because '.' isn't illegal
        # in field names
        lookup = inst_name + "__M_UVM__" + field_name

        r = None  # uvm_resource#(T) r;
        if not pool.exists(lookup):
            r = UVMResource(field_name, inst_name)
            pool.add(lookup, r)
        else:
            r = pool.get(lookup)
            exists = True

        if curr_phase is not None and curr_phase.get_name() == "build":
            r.precedence = UVMResourceBase.default_precedence - cntxt.get_depth()
        else:
            r.precedence = UVMResourceBase.default_precedence

        r.write(value, cntxt)

        if exists:
            rp = UVMResourcePool.get()
            rp.set_priority_name(r, PRI_HIGH)
        else:
            # Doesn't exist yet, so put it in resource db at the head.
            r.set_override()

        # trigger any waiters
        if field_name in UVMConfigDb.m_waiters:
            w = None
            for i in range(len(UVMConfigDb.m_waiters[field_name])):
                w = UVMConfigDb.m_waiters[field_name].get(i)
                if uvm_re_match(uvm_glob_to_re(inst_name),w.inst_name) == 0:
                    w.trigger.set()

        if p is not None:
            p.set_randstate(rstate)

        if UVMConfigDbOptions.is_tracing():
            UVMResourceDb.m_show_msg("CFGDB/SET", "Configuration","set", inst_name, field_name, cntxt, r)

    @classmethod
    def exists(cls, cntxt, inst_name, field_name, spell_chk=False):
        """
        Check if a value for `field_name` is available in `inst_name`, using
        component `cntxt` as the starting search point. `inst_name` is an explicit
        instance name relative to `cntxt` and may be an empty string if the
        `cntxt` is the instance that the configuration object applies to.
        `field_name` is the specific field in the scope that is being searched for.
        The `spell_chk` arg can be set to 1 to turn spell checking on if it
        is expected that the field should exist in the database. The function
        returns 1 if a config parameter exists and 0 if it doesn't exist.

        Args:
            cntxt:
            inst_name:
            field_name:
            spell_chk:
        Returns:
        """
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        if cntxt is None:
            cntxt = cs.get_root()
        if inst_name == "":
            inst_name = cntxt.get_full_name()
        elif cntxt.get_full_name() != "":
            inst_name = cntxt.get_full_name() + "." + inst_name
        found_val = UVMResourceDb.get_by_name(inst_name,field_name,spell_chk)
        return found_val is not None

    #  // Function: wait_modified
    #  //
    #  // Wait for a configuration setting to be set for ~field_name~
    #  // in ~cntxt~ and ~inst_name~. The task blocks until a new configuration
    #  // setting is applied that effects the specified field.
    #
    #  static task wait_modified(UVMComponent cntxt, string inst_name,
    #      string field_name);
    #    process p = process::self();
    #    string rstate = p.get_randstate();
    #    m_uvm_waiter waiter;
    #    uvm_coreservice_t cs = uvm_coreservice_t::get();
    #
    #    if(cntxt == null)
    #      cntxt = cs.get_root();
    #    if(cntxt != cs.get_root()) begin
    #      if(inst_name != "")
    #        inst_name = {cntxt.get_full_name(),".",inst_name};
    #      else
    #        inst_name = cntxt.get_full_name();
    #    end
    #
    #    waiter = new(inst_name, field_name);
    #
    #    if(!m_waiters.exists(field_name))
    #      m_waiters[field_name] = new;
    #    m_waiters[field_name].push_back(waiter);
    #
    #    p.set_randstate(rstate);
    #
    #    // wait on the waiter to trigger
    #    @waiter.trigger;
    #
    #    // Remove the waiter from the waiter list
    #    for(int i=0; i<m_waiters[field_name].size(); ++i) begin
    #      if(m_waiters[field_name].get(i) == waiter) begin
    #        m_waiters[field_name].delete(i);
    #        break;
    #      end
    #    end
    #  endtask
    #
    #
    #endclass

#// Section: Types
#
#//----------------------------------------------------------------------
#// Topic: uvm_config_int
#//
#// Convenience type for uvm_config_db#(uvm_bitstream_t)
#//
#//| typedef uvm_config_db#(uvm_bitstream_t) uvm_config_int;
#typedef uvm_config_db#(uvm_bitstream_t) uvm_config_int;
#
#//----------------------------------------------------------------------
#// Topic: uvm_config_string
#//
#// Convenience type for uvm_config_db#(string)
#//
#//| typedef uvm_config_db#(string) uvm_config_string;
#typedef uvm_config_db#(string) uvm_config_string;
#
#//----------------------------------------------------------------------
#// Topic: uvm_config_object
#//
#// Convenience type for uvm_config_db#(uvm_object)
#//
#//| typedef uvm_config_db#(uvm_object) uvm_config_object;
#typedef uvm_config_db#(uvm_object) uvm_config_object;
#
#//----------------------------------------------------------------------
#// Topic: uvm_config_wrapper
#//
#// Convenience type for uvm_config_db#(uvm_object_wrapper)
#//
#//| typedef uvm_config_db#(uvm_object_wrapper) uvm_config_wrapper;
#typedef uvm_config_db#(uvm_object_wrapper) uvm_config_wrapper;


class UVMConfigDbOptions:
    """
     Class: UVMConfigDbOptions

     Provides a namespace for managing options for the
     configuration DB facility.  The only thing allowed in this class is static
     local data members and static functions for manipulating and
     retrieving the value of the data members.  The static local data
     members represent options and settings that control the behavior of
     the configuration DB facility.

     Options include:

      * tracing:  on/off

        The default for tracing is off.
    """

    #  static local bit ready;
    ready = False
    #  static local bit tracing;
    tracing = False

    @classmethod
    def turn_on_tracing(cls):
        """
        Turn tracing on for the configuration database. This causes all
        reads and writes to the database to display information about
        the accesses. Tracing is off by default.

        This method is implicitly called by the `+UVM_CONFIG_DB_TRACE`.
        """
        if not UVMConfigDbOptions.ready:
            UVMConfigDbOptions.init()
        UVMConfigDbOptions.tracing = True

    @classmethod
    def turn_off_tracing(cls):
        """
        Turn tracing off for the configuration database.
        """
        if not UVMConfigDbOptions.ready:
            UVMConfigDbOptions.init()
        UVMConfigDbOptions.tracing = False

    @classmethod
    def is_tracing(cls):
        """
        Returns: 1 if the tracing facility is on and 0 if it is off.
        """
        if not UVMConfigDbOptions.ready:
            UVMConfigDbOptions.init()
        return UVMConfigDbOptions.tracing


    @classmethod
    def init(cls):
        trace_args = []  # string trace_args[$];
        from .uvm_cmdline_processor import UVMCmdlineProcessor
        clp = UVMCmdlineProcessor.get_inst()
        if clp.get_arg_matches("+UVM_CONFIG_DB_TRACE", trace_args) > 0:
            UVMConfigDbOptions.tracing = 1
        if clp.get_arg_matches("+UVM_CONFIG_DB_TRACE=", trace_args) > 0:
            UVMConfigDbOptions.tracing = 1
        UVMConfigDbOptions.ready = 1



#------------------------------------------------------------------------------
#   Copyright 2011 Mentor Graphics Corporation
#   Copyright 2011 Cadence Design Systems, Inc.
#   Copyright 2011 Synopsys, Inc.
#   Copyright 2013 NVIDIA Corporation
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------
"""
Group: Command Line Debug

This section lists most of the plusargs available for command line usage. Note
that when used with `cocotb`, arguments are usually passed with SIM_ARGS="...".
`cocotb` also supports PLUSARGS="...", so that can be used instead.
For example::

    SIM=icarus make SIM_ARGS="+UVM_TESTNAME=reg_test"

Variable: +UVM_DUMP_CMDLINE_ARGS

~+UVM_DUMP_CMDLINE_ARGS~ allows the user to dump all command line arguments to the
reporting mechanism.  The output in is tree format.

The implementation of this is in `UVMRoot`.

Group: Built-in UVM Aware Command Line Arguments

Variable: +UVM_TESTNAME

~+UVM_TESTNAME=<class name>~ allows the user to specify which uvm_test (or
uvm_component) should be created via the factory and cycled through the UVM phases.
If multiple of these settings are provided, the first occurrence is used and a warning
is issued for subsequent settings.  For example::

  <sim command> +UVM_TESTNAME=read_modify_write_test


The implementation of this is in uvm_root since this is procedurally invoked via
`UVMRoot.run_test`.

Variable: +UVM_VERBOSITY

~+UVM_VERBOSITY=<verbosity>~ allows the user to specify the initial verbosity
for all components.  If multiple of these settings are provided, the first occurrence
is used and a warning is issued for subsequent settings.  For example::

  <sim command> +UVM_VERBOSITY=UVM_HIGH


The implementation of this is in `UVMRoot` since this is procedurally invoked via
`UVMRoot.__init__`.

Variable: +uvm_set_verbosity

~+uvm_set_verbosity=<comp>,<id>,<verbosity>,<phase>~ and
~+uvm_set_verbosity=<comp>,<id>,<verbosity>,time,<time>~ allow the users to manipulate the
verbosity of specific components at specific phases (and times during the "run" phases)
of the simulation.  The ~id~ argument can be either ~_ALL_~ for all IDs or a
specific message id.  Wildcarding is not supported for ~id~ due to performance concerns.
Settings for non-"run" phases are executed in order of occurrence on the command line.
Settings for "run" phases (times) are sorted by time and then executed in order of
occurrence for settings of the same time.  For example::

  <sim command> +uvm_set_verbosity=uvm_test_top.env0.agent1.*,_ALL_,UVM_FULL,time,800


Variable: +uvm_set_action

~+uvm_set_action=<comp>,<id>,<severity>,<action>~ provides the equivalent of
various uvm_report_object's set_report_*_action APIs.  The special keyword,
~_ALL_~, can be provided for both/either the ~id~ and/or ~severity~ arguments.  The
action can be UVM_NO_ACTION or a | separated list of the other UVM message
actions.  For example::

  <sim command> +uvm_set_action=uvm_test_top.env0.*,_ALL_,UVM_ERROR,UVM_NO_ACTION


Variable: +uvm_set_severity

``+uvm_set_severity=<comp>,<id>,<current severity>,<new severity>`` provides the
equivalent of the various uvm_report_object's set_report_*_severity_override APIs. The
special keyword, ~_ALL_~, can be provided for both/either the ~id~ and/or
~current severity~ arguments.  For example::

  <sim command> +uvm_set_severity=uvm_test_top.env0.*,BAD_CRC,UVM_ERROR,UVM_WARNING


Variable: +UVM_TIMEOUT

~+UVM_TIMEOUT=<timeout>,<overridable>~ allows users to change the global timeout of the UVM
framework.  The <overridable> argument ('YES' or 'NO') specifies whether user code can subsequently
change this value.  If set to 'NO' and the user code tries to change the global timeout value, an
warning message will be generated. For example::

  <sim command> +UVM_TIMEOUT=200000,NO


The implementation of this is in uvm_root.

Variable: +UVM_MAX_QUIT_COUNT

~+UVM_MAX_QUIT_COUNT=<count>,<overridable>~ allows users to change max quit count for the report
server.  The <overridable> argument ('YES' or 'NO') specifies whether user code can subsequently
change this value.  If set to 'NO' and the user code tries to change the max quit count value, an
warning message will be generated. For example::

  <sim command> +UVM_MAX_QUIT_COUNT=5,NO


Variable: +UVM_PHASE_TRACE

~+UVM_PHASE_TRACE~ turns on tracing of phase executions.  Users simply need to put the
argument on the command line.

Variable: +UVM_OBJECTION_TRACE

~+UVM_OBJECTION_TRACE~ turns on tracing of objection activity.  Users simply need to put the
argument on the command line.

Variable: +UVM_RESOURCE_DB_TRACE

~+UVM_RESOURCE_DB_TRACE~ turns on tracing of resource DB access.
Users simply need to put the argument on the command line.

Variable: +UVM_CONFIG_DB_TRACE

~+UVM_CONFIG_DB_TRACE~ turns on tracing of configuration DB access.
Users simply need to put the argument on the command line.

Variable: +uvm_set_inst_override

Variable: +uvm_set_type_override

~+uvm_set_inst_override=<req_type>,<override_type>,<full_inst_path>~ and
~+uvm_set_type_override=<req_type>,<override_type>[,<replace>]~ work
like the name based overrides in the factory--factory.set_inst_override_by_name()
 and factory.set_type_override_by_name().
For uvm_set_type_override, the third argument is 0 or 1 (the default is
1 if this argument is left off); this argument specifies whether previous
type overrides for the type should be replaced.  For example::

  <sim command> +uvm_set_type_override=eth_packet,short_eth_packet


The implementation of this is in uvm_root.

Variable: +uvm_set_config_int

Variable: +uvm_set_config_string

~+uvm_set_config_int=<comp>,<field>,<value>~ and
~+uvm_set_config_string=<comp>,<field>,<value>~ work like their
procedural counterparts: set_config_int() and set_config_string(). For
the value of int config settings, 'b (0b), 'o, 'd, 'h ('x or 0x)
as the first two characters of the value are treated as base specifiers
for interpreting the base of the number. Size specifiers are not used
since SystemVerilog does not allow size specifiers in string to
value conversions.  For example::

  <sim command> +uvm_set_config_int=uvm_test_top.soc_env,mode,5

No equivalent of set_config_object() exists since no way exists to pass a
`UVMObject` into the simulation via the command line.


The implementation of this is in uvm_root.

Variable: +uvm_set_default_sequence

The ~+uvm_set_default_sequence=<seqr>,<phase>,<type>~ plusarg allows
the user to define a default sequence from the command line, using the
~typename~ of that sequence.  For example::

  <sim command> +uvm_set_default_sequence=path.to.sequencer,main_phase,seq_type

This is functionally equivalent to calling the following in your
test:

.. code-block:: python
  cs = UVMCoreService.get()
  f = cs.get_factory()
  UVMConfigDb.set(self, "path.to.sequencer.main_phase", "default_sequence",
                                          f.find_wrapper_by_name("seq_type"))

The implementation of this is in `UVMRoot`.
"""


import regex
import cocotb
from typing import List, Optional

from .uvm_report_object import UVMReportObject
from .uvm_debug import uvm_debug
from ..macros import uvm_error
from .uvm_object_globals import UVM_DEBUG, UVM_FULL, UVM_HIGH, UVM_LOW, UVM_MEDIUM, UVM_NONE
from .uvm_globals import uvm_check_output_args


class UVMCmdLineVerb:
    def __init__(self):
        self.comp_path = ""
        self.id = ""
        self.verb = 0
        self.exec_time = 0


def uvm_dpi_get_tool_name():
    """
    Return the simulator name, as reported by VPI

    Returns:
        str: Simulator name.
    """
    return cocotb.SIM_NAME


def uvm_dpi_get_tool_version():
    """
    Return the simulator version, as reported by VPI

    """
    return cocotb.SIM_VERSION


def uvm_dpi_regcomp(match):
    """
    Tries to recompile given string into regular expression.

    Returns:
        Regex|None:
    """
    rr = None
    try:
        rr = regex.compile(match)
    except Exception as e:
        print(str(e))
    return rr


def uvm_dpi_regexec(exp_h, elem):
    if exp_h.match(elem) is not None:
        return 0
    return 1


class UVMCmdlineProcessor(UVMReportObject):
    """
    This class provides an interface to the command line arguments that
    were provided for the given simulation.  The class is intended to be
    used as a singleton, but that isn't required.  The generation of the
    data structures which hold the command line argument information
    happens during construction of the class object.  A global variable
    called ~uvm_cmdline_proc~ is created at initialization time and may
    be used to access command line information.

    The uvm_cmdline_processor class also provides support for setting various UVM
    variables from the command line such as components' verbosities and configuration
    settings for integral types and strings.  Each of these capabilities is described
    in the Built-in UVM Aware Command Line Arguments section.
    """


    m_inst: Optional['UVMCmdlineProcessor'] = None
    uvm_cmdline_proc = None

    # Used in unit tests only (not part of original UVM)
    m_test_mode = False
    m_test_plusargs = {}
    m_test_argv = []

    # Group: Singleton

    @classmethod
    def get_inst(cls) -> 'UVMCmdlineProcessor':
        """
        Returns the singleton instance of the UVM command line processor.

        Returns:
            UVMCmdlineProcessor: Singleton instance of UVM command line processor.
        """
        if UVMCmdlineProcessor.m_inst is None:
            UVMCmdlineProcessor.m_inst = UVMCmdlineProcessor("uvm_cmdline_proc")
        return UVMCmdlineProcessor.m_inst

    # Group: Basic Arguments

    def get_args(self) -> List[str]:
        """
        This function returns a queue with all of the command line
        arguments that were used to start the simulation. Note that
        element 0 of the array will always be the name of the
        executable which started the simulation.

        Returns:
            list: Queue of all the command line arguments.
        """
        return self.m_argv

    def get_plusargs(self):
        """
        This function returns a list with all of the plus arguments
        that were used to start the simulation. Plusarguments may be
        used by the simulator vendor, or may be specific to a company
        or individual user. Plusargs never have extra arguments
        (i.e. if there is a plusarg as the second argument on the
        command line, the third argument is unrelated); this is not
        necessarily the case with vendor specific dash arguments.

        Returns:
            list: List of all used plusargs.
        """
        return self.m_plus_argv

    def get_uvm_args(self):
        """
        This function returns a queue with all of the uvm arguments
        that were used to start the simulation. A UVM argument is
        taken to be any argument that starts with a - or + and uses
        the keyword UVM (case insensitive) as the first three
        letters of the argument.

        Returns:
            list: List of used UVM arguments, including name and value
        """
        return self.m_uvm_argv

    def get_arg_matches(self, match, args):
        """
        This function loads a queue with all of the arguments that
        match the input expression and returns the number of items
        that matched. If the input expression is bracketed
        with #, then it is taken as an extended regular expression
        otherwise, it is taken as the beginning of an argument to match.
        For example:

        .. code-block:: python

          myargs = []
            uvm_cmdline_proc.get_arg_matches("+foo",myargs))  #matches +foo, +foobar
                                                              #doesn't match +barfoo
            uvm_cmdline_proc.get_arg_matches("/foo/",myargs)) #matches +foo, +foobar,
                                                              #foo.sv, barfoo, etc.
            uvm_cmdline_proc.get_arg_matches("/^foo.*\\.sv",myargs)) #matches foo.sv
                                                                    #and foo123.sv,
                                                                    #not barfoo.sv.
        Args:
            match (str): String to match.
            args (list): List into which matches are appended.

        Returns:
            int: Number of matches found.
        """
        exp_h = None
        match_len = len(match)
        args.clear()
        if len(match) > 2 and match[0] == "/" and match[len(match)-1] == "/":
            match = match[1:len(match)-2]
            exp_h = uvm_dpi_regcomp(match)
            if exp_h is None:
                uvm_error("UVM_CMDLINE_PROC",
                        "Unable to compile the regular expression: " + match)
                return 0
        for i in range(len(self.m_argv)):
            if exp_h is not None:
                if uvm_dpi_regexec(exp_h, self.m_argv[i]) == 0:
                    args.append(self.m_argv[i])
            elif len(self.m_argv[i]) >= match_len:
                sub_argv = self.m_argv[i][0:match_len]
                if sub_argv == match:
                    args.append(self.m_argv[i])
        return len(args)

    # Group: Argument Values

    def get_arg_value(self, match, value):
        """
        This function finds the first argument which matches the `match` arg and
        returns the suffix of the argument. This is similar to the
        `sv.value_plusargs` function, but does not take a formatting string. The
        return value is the number of command line arguments that match the `match` string,
        and `value` is the value of the first match.

        Args:
            match (str): String/pattern to match.
            value (list): List into which matches are appended.
        Returns:
            int: Number of cmd line args matching given match.
        """
        uvm_check_output_args([value])
        chars = len(match)
        get_arg_value = 0
        for i in range(0, len(self.m_argv)):
            if len(self.m_argv[i]) >= chars:
                arg_to_try = self.m_argv[i][0:chars]

                if arg_to_try == match:
                    get_arg_value += 1
                    if get_arg_value == 1:
                        value.append(self.m_argv[i][chars:len(self.m_argv[i])])
        return get_arg_value

    def get_arg_values(self, match, values):
        """
        This function finds all the arguments which matches the `match` arg and
        returns the suffix of the arguments in a list of values. The return
        value is the number of matches that were found (it is the same as
        values.size()).
        For example if '+foo=1,yes,on +foo=5,no,off' was provided on the command
        line and the following code was executed:

        .. code-block:: python

          foo_values = []
          cmd_line_proc = UVMCmdlineProcessor()
          cmd_line_proc.get_arg_values("+foo=", foo_values)


        The foo_values queue would contain two entries.  These entries are shown
        here::

          0 - "1,yes,on"
          1 - "5,no,off"

        Splitting the resultant string is left to user but using the
        uvm_split_string() function is recommended.
        Args:
            match (str): Argument to be matched.
            values (list): Values of matching arguments.
        Returns:
            int: Number of arguments that match.
        """
        values.clear()
        #if match[0] == '+':
        #    match = match[1:]
        #if match[-1] == '=':
        #    match = match[:-1]

        chars = len(match)
        for i in range(len(self.m_argv)):
            uvm_debug(self, 'get_arg_values', 'Checking ' + self.m_argv[i] + ' - '
                + match)
            if len(self.m_argv[i]) >= chars:
                argv_str = self.m_argv[i][0:chars]
                if argv_str == match:
                    val = self.m_argv[i][chars:len(self.m_argv[i])]
                    values.append(val)
            else:
                pass
        return len(values)

    # Group: Tool information

    def get_tool_name(self):
        """
        Returns the simulation tool that is executing the simulation.
        This is a vendor specific string.

        Returns:
            str: Toolname.
        """
        return uvm_dpi_get_tool_name()

    def get_tool_version(self):
        """
        Returns the version of the simulation tool that is executing the simulation.
        This is a vendor specific string.

        Returns:
            str: Tool version.
        """
        return uvm_dpi_get_tool_version()

    def __init__(self, name=""):
        """
        Constructor which handles also cmdline argument processing.

        Args:
            name (str): Name of this object.
        """
        super().__init__(name)
        self.m_argv = []
        self.m_plus_argv = []
        self.m_uvm_argv = []


        argv = []
        if hasattr(cocotb, 'argv') and cocotb.argv is not None:
            argv = cocotb.argv
        else:
            import sys
            argv = sys.argv
        if UVMCmdlineProcessor.m_test_mode is True:
            argv = UVMCmdlineProcessor.m_test_argv
        if argv is not None:
            self.extract_args(argv)


    def extract_args(self, argv):
        """
        Extracts simulation arguments from given array. Stores results into 3
        different internal arrays: All args, plusargs and UVM args.

        Args:
            argv (list): List of arguments.
        """
        sub = ""
        for arg in argv:
            if len(arg) == 0:
                continue
            self.m_argv.append(arg)
            if arg[0] == "+":
                self.m_plus_argv.append(arg[0])
                uvm_debug(self, '__init__', "Simple UVM plusarg %s" % (arg))
            sub = arg[1:4]
            sub = sub.upper()
            if sub == "UVM":
                self.m_uvm_argv.append(arg)
                uvm_debug(self, '__init__', "Found UVM plusarg %s" % (arg))


    def m_convert_verb(self, verb_str):
        """
        Converts given verbosity string into a number representing that
        verbosity level.

        Args:
            verb_str (str): Verbosity string.
        Returns:
            int: Verbosity level matching one in `UVM_VERBOSITY_LIST`.
        """
        if verb_str == "NONE":
            return UVM_NONE
        if verb_str == "UVM_NONE":
            return UVM_NONE
        if verb_str == "LOW":
            return UVM_LOW
        if verb_str == "UVM_LOW":
            return UVM_LOW
        if verb_str == "MEDIUM":
            return UVM_MEDIUM
        if verb_str == "UVM_MEDIUM":
            return UVM_MEDIUM
        if verb_str == "HIGH":
            return UVM_HIGH
        if verb_str == "UVM_HIGH":
            return UVM_HIGH
        if verb_str == "FULL":
            return UVM_FULL
        if verb_str == "UVM_FULL":
            return UVM_FULL
        if verb_str == "DEBUG":
            return UVM_DEBUG
        if verb_str == "UVM_DEBUG":
            return UVM_DEBUG
        return -1

UVMCmdlineProcessor.uvm_cmdline_proc = UVMCmdlineProcessor.get_inst()



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2008 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2019 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb

from ...macros import uvm_object_utils, uvm_info, uvm_error
from .. import UVMRegSequence
from ...base.uvm_resource_db import UVMResourceDb
from ...base.sv import sv
from ...base import UVM_LOW, UVM_HIGH
from ...base.uvm_globals import uvm_zero_delay
from ..uvm_reg_model import UVM_CHECK, UVM_FRONTDOOR, UVM_IS_OK


class UVMRegHWResetSeq(UVMRegSequence):
    """

     class: UVMRegHWResetSeq
     Test the hard reset values of registers

     The test sequence performs the following steps

     1. resets the DUT and the
     block abstraction class associated with this sequence.

     2. reads all of the registers in the block,
     via all of the available address maps,
     comparing the value read with the expected reset value.

     If bit-type resource named
     "NO_REG_TESTS" or "NO_REG_HW_RESET_TEST"
     in the "REG::" namespace
     matches the full name of the block or register,
     the block or register is not tested.

     .. code-block:: python
         UVMResourceDb.set("REG::" + regmodel.blk.get_full_name() + ".*",
                               "NO_REG_TESTS", 1, self)

    This is usually the first test executed on any DUT.
    """


    def __init__(self, name="UVMRegHWResetSeq"):
        super().__init__(name)
        #   // Variable: model
        #   //
        #   // The block to be tested. Declared in the base class.
        #   //
        self.model = None  # uvm_reg_block


    #   // Variable: body
    #   //
    #   // Executes the Hardware Reset sequence.
    #   // Do not call directly. Use seq.start() instead.
    #
    
    async def body(self):
        if self.model is None:
            uvm_error("UVMRegHWResetSeq", "Not block or system specified to run sequence on")
            return
        uvm_info("STARTING_SEQ", "|UVMRegHWResetSeq| Starting " + self.get_name() +
                " sequence...",UVM_LOW)

        await self.reset_blk(self.model)
        self.model.reset()

        await self.do_block(self.model)
    #   endtask: body

    #// Task: do_block
    #   //
    #   // Test all of the registers in a given ~block~
    #   //
    #   protected virtual task do_block(uvm_reg_block blk)
    
    async def do_block(self, blk):
        maps = []  # uvm_reg_map maps[$]
        sub_maps = []  # uvm_reg_map sub_maps[$]
        no_rr1 = UVMResourceDb.get_by_name("REG::" + blk.get_full_name(), "NO_REG_TESTS", 0)
        no_rr2 = UVMResourceDb.get_by_name("REG::" + blk.get_full_name(), "NO_REG_HW_RESET_TEST", 0)

        if no_rr1 is not None or no_rr2 is not None:
            return

        blk.get_maps(maps)

        # Iterate over all maps defined for the RegModel block
        for d in range(len(maps)):
            regs = []  # uvm_reg[]
            maps[d].get_submaps(sub_maps)
            if len(sub_maps) != 0:
                continue

            # Iterate over all registers in the map, checking accesses
            # Note: if map were in inner loop, could test simulataneous
            # access to same reg via different bus interfaces

            regs = []
            maps[d].get_registers(regs)

            for i in range(len(regs)):
                status = 0
                reg_name = regs[i].get_full_name()

                # Registers with certain attributes are not to be tested
                dont_test1 = UVMResourceDb.get_by_name("REG::" + reg_name, "NO_REG_HW_RESET_TEST", 0)
                dont_test2 = UVMResourceDb.get_by_name("REG::" + reg_name, "NO_REG_TESTS", 0)
                if (dont_test1 is not None or dont_test2 is not None):
                    continue

                uvm_info(self.get_type_name(),
                        sv.sformatf("Verifying reset value of register %s in map \"%s\"...",
                            regs[i].get_full_name(), maps[d].get_full_name()), UVM_LOW)

                status = []
                await regs[i].mirror(status, UVM_CHECK, UVM_FRONTDOOR, maps[d], self)
                status = status[0]

                if status != UVM_IS_OK:
                    uvm_error(self.get_type_name(),
                          sv.sformatf("Status was %s when reading reset value of register \"%s\" through map \"%s\".",
                          status.name(), regs[i].get_full_name(), maps[d].get_full_name()))

            blks = []  # uvm_reg_block blks[$]
            blk.get_blocks(blks)
            for i in range(len(blks)):
                await self.do_block(blks[i])

        #   endtask:do_block

    #   // task: reset_blk
    #   // Reset the DUT that corresponds to the specified block abstraction class.
    #   //
    #   // Currently empty.
    #   // Will rollback the environment's phase to the ~reset~
    #   // phase once the new phasing is available.
    #   //
    #   // In the meantime, the DUT should be reset before executing this
    #   // test sequence or this method should be implemented
    #   // in an extension to reset the DUT.
    #   //
    #   virtual task reset_blk(uvm_reg_block blk)
    
    async def reset_blk(self, blk):
        await uvm_zero_delay()

    #endclass: UVMRegHWResetSeq
uvm_object_utils(UVMRegHWResetSeq)



import cocotb
from cocotb.triggers import Timer
from uvm import *

class NewTest(UVMTest):
    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        await Timer(100, "NS")
        phase.drop_objection(self)

uvm_component_utils(NewTest)

class NewTest2(NewTest):
    pass

@cocotb.test()
async def test_dut(dut):
    await run_test('NewTest')

# Cannot run 2 tests with same run yet
#@cocotb.test()
#async def test_dut2(dut):
#    await run_test('NewTest2')




from .uvm_global_defines import *
from .uvm_message_defines import *
from .uvm_object_defines import *
from .uvm_sequence_defines import *
from .uvm_reg_defines import *
from .uvm_callback_defines import *



#//----------------------------------------------------------------------
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2011 Mentor Graphics Corporation
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
#// CLASS: uvm_tlm_time
#// Canonical time type that can be used in different timescales
#//
#// This time type is used to represent time values in a canonical
#// form that can bridge initiators and targets located in different
#// timescales and time precisions.
#//
#// For a detailed explanation of the purpose for this class,
#// see <Why is this necessary>.
#//
from uvm.macros.uvm_message_defines import uvm_error, uvm_fatal
from uvm.base.sv import cat
class UVMTLMTime():
#
#   static local real m_resolution = 1.0e-12; // ps by default
    m_resolution = 1.0e-12 
#   local real m_res;
#   local time m_time;  // Number of 'm_res' time units,
#   local string m_name;
#
#   // Function: set_time_resolution
#   // Set the default canonical time resolution.
#   //
#   // Must be a power of 10.
#   // When co-simulating with SystemC, it is recommended
#   // that default canonical time resolution be set to the
#   // SystemC time resolution.
#   //
#   // By default, the default resolution is 1.0e-12 (ps)
#   //
    @staticmethod
    def set_time_resolution(res):
        UVMTLMTime.m_resolution = res

#   // Function: new
#   // Create a new canonical time value.
#   //
#   // The new value is initialized to 0.
#   // If a resolution is not specified,
#   // the default resolution,
#   // as specified by <set_time_resolution()>,
#   // is used.
    def __init__(self, name="uvm_tlm_time", res=0):
        self.m_name = name
        self.m_res = UVMTLMTime.m_resolution if (res == 0) else res
        self.reset()

#   // Function: get_name
#   // Return the name of this instance
#   //
    def get_name(self):
        return self.m_name

#   // Function: reset
#   // Reset the value to 0
    def reset(self):
        self.m_time = 0


#   // Scale a timescaled value to 'm_res' units,
#   // the specified scale
    def _to_m_res(self, t, scaled, secs):
#      // ToDo: Check resolution
        return (t/scaled) * (secs/self.m_res)

#   // Function: get_realtime
#   // Return the current canonical time value,
#   // scaled for the caller's timescale
#   //
#   // ~scaled~ must be a time literal value that corresponds
#   // to the number of seconds specified in ~secs~ (1ns by default).
#   // It must be a time literal value that is greater or equal
#   // to the current timescale.
#   //
#   //| #(delay.get_realtime(1ns));
#   //| #(delay.get_realtime(1fs, 1.0e-15));
#   //
    def get_realtime(self, scaled, secs = 1.0e-9):
        return self.m_time*scaled * (self.m_res/secs)

#   // Function: incr
#   // Increment the time value by the specified number of scaled time unit
#   //
#   // ~t~ is a time value expressed in the scale and precision
#   // of the caller.
#   // ~scaled~ must be a time literal value that corresponds
#   // to the number of seconds specified in ~secs~ (1ns by default).
#   // It must be a time literal value that is greater or equal
#   // to the current timescale.
#   //
#   //| delay.incr(1.5ns, 1ns);
#   //| delay.incr(1.5ns, 1ps, 1.0e-12);
#   //
    def incr(self, t, scaled, secs = 1.0e-9):
        if (t < 0.0):
            uvm_error("UVM/TLM/TIMENEG", cat("Cannot increment uvm_tlm_time variable ", self.m_name, " by a negative value"))
            return
        if (scaled == 0):
            uvm_fatal("UVM/TLM/BADSCALE",
                    "uvm_tlm_time::incr() called with a scaled time literal that is smaller than the current timescale")

        self.m_time += self._to_m_res(t, scaled, secs)

#   // Function: decr
#   // Decrement the time value by the specified number of scaled time unit
#   //  
#   // ~t~ is a time value expressed in the scale and precision
#   // of the caller.
#   // ~scaled~ must be a time literal value that corresponds
#   // to the number of seconds specified in ~secs~ (1ns by default).
#   // It must be a time literal value that is greater or equal
#   // to the current timescale.
#   //
#   //| delay.decr(200ps, 1ns);
#   //
    def decr(self, t, scaled, secs):
        if (t < 0.0):
            uvm_error("UVM/TLM/TIMENEG", cat("Cannot decrement uvm_tlm_time variable ", self.m_name, " by a negative value"));
            return
        if (scaled == 0):
            uvm_fatal("UVM/TLM/BADSCALE",
                 "uvm_tlm_time::decr() called with a scaled time literal that is smaller than the current timescale")
      
        self.m_time -= self._to_m_res(t, scaled, secs);

        if (self.m_time < 0.0):
            uvm_error("UVM/TLM/TOODECR", cat("Cannot decrement uvm_tlm_time variable ", self.m_name, " to a negative value"))
            self.reset()

#   // Function: get_abstime
#   // Return the current canonical time value,
#   // in the number of specified time unit, regardless of the
#   // current timescale of the caller.
#   //
#   // ~secs~ is the number of seconds in the desired time unit
#   // e.g. 1e-9 for nanoseconds.
#   //
#   //| $write("%.3f ps\n", delay.get_abstime(1e-12));
#   //
    def get_abstime(self, secs):
        return self.m_time*self.m_res/secs

#   // Function: set_abstime
#   // Set the current canonical time value,
#   // to the number of specified time unit, regardless of the
#   // current timescale of the caller.
#   //
#   // ~secs~ is the number of seconds in the time unit in the value ~t~
#   // e.g. 1e-9 for nanoseconds.
#   //
#   //| delay.set_abstime(1.5, 1e-12));
#   //
    def set_abstime(self, t, secs):
        self.m_time = t*secs/self.m_res

#// Group: Why is this necessary
#//
#// Integers are not sufficient, on their own,
#// to represent time without any ambiguity:
#// you need to know the scale of that integer value.
#// That scale is information conveyed outside of that integer.
#// In SystemVerilog, it is based on the timescale
#// that was active when the code was compiled.
#// SystemVerilog properly scales time literals, but not integer values.
#// That's because it does not know the difference between an integer
#// that carries an integer value and an integer that carries a time value.
#// The 'time' variables are simply 64-bit integers,
#// they are not scaled back and forth to the underlying precision.
#//
#//| `timescale 1ns/1ps
#//| 
#//| module m();
#//| 
#//| time t;
#//| 
#//| initial
#//| begin
#//|    #1.5;
#//|    $write("T=%f ns (1.5)\n", $realtime());
#//|    t = 1.5;
#//|    #t;
#//|    $write("T=%f ns (3.0)\n", $realtime());
#//|    #10ps;
#//|    $write("T=%f ns (3.010)\n", $realtime());
#//|    t = 10ps;
#//|    #t;
#//|    $write("T=%f ns (3.020)\n", $realtime());
#//| end
#//| endmodule
#//
#// yields
#//
#//| T=1.500000 ns (1.5)
#//| T=3.500000 ns (3.0)
#//| T=3.510000 ns (3.010)
#//| T=3.510000 ns (3.020)
#//
#// Within SystemVerilog, we have to worry about
#// - different time scale
#// - different time precision
#//
#// Because each endpoint in a socket
#// could be coded in different packages
#// and thus be executing under different timescale directives,
#// a simple integer cannot be used to exchange time information
#// across a socket.
#//
#// For example
#//
#//| `timescale 1ns/1ps 
#//| 
#//| package a_pkg; 
#//| 
#//| class a; 
#//|    function void f(inout time t); 
#//|       t += 10ns; 
#//|    endfunction
#//| endclass 
#//| 
#//| endpackage 
#//| 
#//| 
#//| `timescale 1ps/1ps 
#//| 
#//| program p; 
#//| 
#//| import a_pkg::*;
#//|
#//| time t;
#//| 
#//| initial
#//| begin 
#//|    a A = new; 
#//|    A.f(t); 
#//|    #t; 
#//|    $write("T=%0d ps (10,000)\n", $realtime());
#//| end
#//| endprogram
#//  
#// yields
#//  
#//| T=10 ps (10,000)
#//
#// Scaling is needed every time you make a procedural call
#// to code that may interpret a time value in a different timescale.
#//
#// Using the uvm_tlm_time type
#//
#//| `timescale 1ns/1ps
#//| 
#//|    package a_pkg;
#//| 
#//| import uvm_pkg::*;
#//| 
#//| class a;
#//|    function void f(uvm_tlm_time t);
#//|       t.incr(10ns, 1ns);
#//|    endfunction
#//| endclass
#//| 
#//| endpackage
#//| 
#//| 
#//| `timescale 1ps/1ps
#//| 
#//| program p;
#//| 
#//| import uvm_pkg::*;
#//| import a_pkg::*;
#//| 
#//| uvm_tlm_time t = new;
#//| 
#//| initial
#//|    begin
#//|       a A = new;
#//|       A.f(t);
#//|       #(t.get_realtime(1ns));
#//|       $write("T=%0d ps (10,000)\n", $realtime());
#//| end
#//| endprogram
#//
#// yields
#//
#//| T=10000 ps (10,000)
#//
#// A similar procedure is required when crossing any simulator
#// or language boundary,
#// such as interfacing between SystemVerilog and SystemC.
#      
# 
# 



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2013-2014 NVIDIA Corporation
#//   Copyright 2019 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from .. import version

uvm_mgc_copyright  = "(C) 2007-2014 Mentor Graphics Corporation"
uvm_cdn_copyright  = "(C) 2007-2014 Cadence Design Systems, Inc."
uvm_snps_copyright = "(C) 2006-2014 Synopsys, Inc."
uvm_cy_copyright   = "(C) 2011-2013 Cypress Semiconductor Corp."
uvm_nv_copyright   = "(C) 2013-2014 NVIDIA Corporation"
uvm_revision = "uvm-python {} (Ported from UVM 1.2)".format(version.__version__)
uvm_tpoikela_copyright = "(C) 2019-2021 Tuomas Poikela (tpoikela)"

def uvm_revision_string():
    return uvm_revision



#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

from .uvm_sequencer_base import UVMSequencerBase
from .uvm_sequencer_analysis_fifo import UVMSequencerAnalysisFIFO
from ..base.uvm_queue import UVMQueue
from ..tlm1.uvm_tlm_fifos import UVMTLMFIFO
from ..tlm1 import (UVMAnalysisExport)
from ..macros.uvm_message_defines import uvm_fatal
from ..base.uvm_globals import *
from ..base.sv import sv

ERR_MSG2 = ("Concurrent calls to get_next_item() not supported. Consider "
    + "using a semaphore to ensure that concurrent processes take turns in the driver")

INFO_MSG1 = ("Dropping response for sequence %0d, sequence not found.  Probable"
    + "cause: sequence exited or has been killed")



class UVMSequencerParamBase(UVMSequencerBase):
    """

    CLASS: uvm_sequencer_param_base #(REQ,RSP)

    Extends <uvm_sequencer_base> with an API depending on specific
    request (REQ) and response (RSP) types.

    Port (UVMAnalysisExport): rsp_export

    Drivers or monitors can connect to this port to send responses
    to the sequencer.  Alternatively, a driver can send responses
    via its seq_item_port::

      seq_item_port.item_done(response)
      seq_item_port.put(response)
      rsp_port.write(response)   <--- via this export

    The rsp_port in the driver and/or monitor must be connected to the
    rsp_export in this sequencer in order to send responses through the
    response analysis port.
    """


    def __init__(self, name, parent):
        """
        Creates and initializes an instance of this class using the normal
        constructor arguments for uvm_component: name is the name of the instance,
        and parent is the handle to the hierarchical parent, if any.

        Args:
            name:
            parent:
        """
        UVMSequencerBase.__init__(self, name, parent)
        self.m_last_req_buffer = UVMQueue()
        self.m_last_rsp_buffer = UVMQueue()
        self.m_num_last_reqs = 1
        self.num_last_items = self.m_num_last_reqs
        self.m_num_last_rsps = 1
        self.m_num_reqs_sent = 0
        self.m_num_rsps_received = 0

        # TODO Adding self as 2nd arg makes the sim fail, it does not advance
        # properly. This is uvm_phase issue most likely
        #self.m_req_fifo = UVMTLMFIFO(name + "__" + "m_req_fifo", None)
        self.m_req_fifo = UVMTLMFIFO(name + "__" + "m_req_fifo", self)  # uvm_tlm_fifo
        self.m_req_fifo.print_enabled = False
        self.rsp_export = UVMAnalysisExport("rsp_export", self)
        self.sqr_rsp_analysis_fifo = UVMSequencerAnalysisFIFO("sqr_rsp_analysis_fifo", self)
        self.sqr_rsp_analysis_fifo.print_enabled = 0


    def send_request(self, sequence_ptr, t, rerandomize=False):
        """
        The send_request function may only be called after a wait_for_grant call.
        This call will send the request item, t,  to the sequencer pointed to by
        sequence_ptr. The sequencer will forward it to the driver. If rerandomize
        is set, the item will be randomized before being sent to the driver.

        Args:
            sequence_ptr (UVMSequenceBase):
            t (UVMSequenceItem):
            rerandomize (bool):
        """
        param_t = None  # REQ

        if sequence_ptr is None:
            self.uvm_report_fatal("SNDREQ", "Send request sequence_ptr is null", UVM_NONE)

        if sequence_ptr.m_wait_for_grant_semaphore < 1:
            self.uvm_report_fatal("SNDREQ", "Send request called without wait_for_grant", UVM_NONE)
        sequence_ptr.m_wait_for_grant_semaphore -= 1

        #if ($cast(param_t, t)):
        param_t = t
        if rerandomize is True:
            if param_t.randomize() is False:
                self.uvm_report_warning("SQRSNDREQ", "Failed to rerandomize sequence item in send_request")
        if param_t.get_transaction_id() == -1:
            param_t.set_transaction_id(sequence_ptr.m_next_transaction_id)
            sequence_ptr.m_next_transaction_id += 1
        self.m_last_req_push_front(param_t)
        #end else begin
        #  uvm_report_fatal(get_name(),$sformatf("send_request failed to cast sequence item"), UVM_NONE)
        #end

        param_t.set_sequence_id(sequence_ptr.m_get_sqr_sequence_id(self.m_sequencer_id, 1))
        t.set_sequencer(self)
        if self.m_req_fifo.try_put(param_t) is False:
            uvm_fatal(self.get_full_name(), ERR_MSG2)

        self.m_num_reqs_sent += 1
        # Grant any locks as soon as possible
        self.grant_queued_locks()


    def get_current_item(self):
        """
        Returns the request_item currently being executed by the sequencer. If the
        sequencer is not currently executing an item, this method will return `null`.

        The sequencer is executing an item from the time that get_next_item or peek
        is called until the time that get or item_done is called.

        Note that a driver that only calls get() will never show a current item,
        since the item is completed at the same time as it is requested.

        Returns:
            UVMSequenceItem: Request item currently being executed
        """
        t = []
        if not self.m_req_fifo.try_peek(t):
            return None
        return t[0]


    def get_num_reqs_sent(self):
        """
        Returns the number of requests that have been sent by this sequencer.

        Returns:
            int: Number of requests sent.
        """
        return self.m_num_reqs_sent


    #  // Function: set_num_last_reqs
    #  //
    #  // Sets the size of the last_requests buffer.  Note that the maximum buffer
    #  // size is 1024.  If max is greater than 1024, a warning is issued, and the
    #  // buffer is set to 1024.  The default value is 1.
    #  //
    #  extern function void set_num_last_reqs(int unsigned max)
    def set_num_last_reqs(self, max_val):
        if max_val > 1024:
            self.uvm_report_warning("HSTOB",
                sv.sformatf("Invalid last size; 1024 is the maximum and will be used"))
            max_val = 1024

        # shrink the buffer if necessary
        while ((self.m_last_req_buffer.size() != 0) and
                (self.m_last_req_buffer.size() > max_val)):
            self.m_last_req_buffer.pop_back()

        self.m_num_last_reqs = max_val
        self.num_last_items = max_val


    def get_num_last_reqs(self):
        """
          Function: get_num_last_reqs

          Returns the size of the last requests buffer, as set by set_num_last_reqs.

         extern function int unsigned get_num_last_reqs()
        Returns:
        """
        return self.m_num_last_reqs


    #  // Function: last_req
    #  //
    #  // Returns the last request item by default.  If n is not 0, then it will get
    #  // the n�th before last request item.  If n is greater than the last request
    #  // buffer size, the function will return ~null~.
    #  //
    #  function REQ last_req(int unsigned n = 0)
    #    if(n > m_num_last_reqs):
    #      uvm_report_warning("HSTOB",
    #        $sformatf("Invalid last access (%0d), the max history is %0d", n,
    #        m_num_last_reqs))
    #      return null
    #    end
    #    if(n == m_last_req_buffer.size())
    #      return null
    #
    #    return m_last_req_buffer[n]
    #  endfunction


    #  //-----------------
    #  // Group: Responses
    #  //-----------------


    def get_num_rsps_received(self):
        """
        Returns the number of responses received thus far by this sequencer.

        Returns:
        """
        return self.m_num_rsps_received



    #  // Function: set_num_last_rsps
    #  //
    #  // Sets the size of the last_responses buffer.  The maximum buffer size is
    #  // 1024. If max is greater than 1024, a warning is issued, and the buffer is
    #  // set to 1024.  The default value is 1.
    #  //
    #  extern function void set_num_last_rsps(int unsigned max)


    def get_num_last_rsps(self):
        """
          Function: get_num_last_rsps

          Returns the max size of the last responses buffer, as set by
          set_num_last_rsps.

         extern function int unsigned get_num_last_rsps()
        Returns:
        """
        return self.m_num_last_rsps


    #  // Function: last_rsp
    #  //
    #  // Returns the last response item by default.  If n is not 0, then it will
    #  // get the nth-before-last response item.  If n is greater than the last
    #  // response buffer size, the function will return ~null~.
    #  //
    #  function RSP last_rsp(int unsigned n = 0)
    #    if(n > m_num_last_rsps):
    #      uvm_report_warning("HSTOB",
    #        $sformatf("Invalid last access (%0d), the max history is %0d", n,
    #        m_num_last_rsps))
    #      return null
    #    end
    #    if(n == m_last_rsp_buffer.size())
    #      return null
    #
    #    return m_last_rsp_buffer[n]
    #  endfunction


    #  // Internal methods and variables; do not use directly, not part of standard


    def m_last_rsp_push_front(self, item):
        """
         /* local */ extern function void m_last_rsp_push_front(RSP item)
        Args:
            item:
        """
        if self.m_num_last_rsps == 0:
            return

        if self.m_last_rsp_buffer.size() == self.m_num_last_rsps:
            self.m_last_rsp_buffer.pop_back()
        #
        self.m_last_rsp_buffer.push_front(item)

    def put_response(self, t):
        """
         /* local */ extern function void put_response (RSP t)
        Args:
            t:
        """
        sequence_ptr = None  # #  uvm_sequence_base

        if t is None:
            self.uvm_report_fatal("SQRPUT", "Driver put a null response", UVM_NONE)

        self.m_last_rsp_push_front(t)
        self.m_num_rsps_received += 1

        # Check that set_id_info was called
        if t.get_sequence_id() == -1:
            #`ifndef CDNS_NO_SQR_CHK_SEQ_ID
            self.uvm_report_fatal("SQRPUT", "Driver put a response with null sequence_id", UVM_NONE)
            #`endif
            return

        sequence_ptr = self.m_find_sequence(t.get_sequence_id())

        if sequence_ptr is not None:
            # If the response_handler is enabled for this sequence, then call the response handler
            if sequence_ptr.get_use_response_handler() == 1:
                sequence_ptr.response_handler(t)
                return
            sequence_ptr.put_response(t)
        else:
            self.uvm_report_info("Sequencer", sv.sformatf(INFO_MSG1, t.get_sequence_id()))
        #endfunction

    def build_phase(self, phase):
        """
         /* local */ extern virtual function void build_phase(uvm_phase phase)
        Args:
            phase:
        """
        super().build_phase(phase)
        self.sqr_rsp_analysis_fifo.sequencer_ptr = self

    def connect_phase(self, phase):
        """
         /* local */ extern virtual function void connect_phase(uvm_phase phase)
        Args:
            phase:
        """
        super().connect_phase(phase)
        self.rsp_export.connect(self.sqr_rsp_analysis_fifo.analysis_export)


    #  /* local */ extern virtual function void do_print (uvm_printer printer)

    #  /* local */ extern virtual function void analysis_write(uvm_sequence_item t)

    def m_last_req_push_front(self, item):
        """
         /* local */ extern function void m_last_req_push_front(REQ item)
        Args:
            item:
        """
        if self.m_num_last_reqs == 0:
            return

        if self.m_last_req_buffer.size() == self.m_num_last_reqs:
            self.m_last_req_buffer.pop_back()

        self.m_last_req_buffer.push_front(item)


#//------------------------------------------------------------------------------
#// IMPLEMENTATION
#//------------------------------------------------------------------------------
#
#// do_print
#// --------
#
#function void uvm_sequencer_param_base::do_print (uvm_printer printer)
#  super.do_print(printer)
#  printer.print_field_int("num_last_reqs", m_num_last_reqs, $bits(m_num_last_reqs), UVM_DEC)
#  printer.print_field_int("num_last_rsps", m_num_last_rsps, $bits(m_num_last_rsps), UVM_DEC)
#endfunction
#
#
#
#// analysis_write
#// --------------
#
#function void uvm_sequencer_param_base::analysis_write(uvm_sequence_item t)
#  RSP response
#
#  if (!$cast(response, t)):
#    uvm_report_fatal("ANALWRT", "Failure to cast analysis port write item", UVM_NONE)
#  end
#  put_response(response)
#endfunction
#
#
#
#
#
#
#// set_num_last_rsps
#// -----------------
#
#function void uvm_sequencer_param_base::set_num_last_rsps(int unsigned max)
#  if(max > 1024):
#    uvm_report_warning("HSTOB",
#      $sformatf("Invalid last size; 1024 is the maximum and will be used"))
#    max = 1024
#  end
#
#  //shrink the buffer
#  while((m_last_rsp_buffer.size() != 0) && (m_last_rsp_buffer.size() > max)):
#    void'(m_last_rsp_buffer.pop_back())
#  end
#
#  m_num_last_rsps = max
#
#endfunction



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from uvm.macros import *
from uvm.reg import UVMRegSequence
from uvm.base import sv
from uvm.reg.uvm_reg_model import *


class user_test_seq(UVMRegSequence):

    def __init__(self, name="user_test_seq"):
        super().__init__(name)
        self.addr = None  # type: int
        self.data = None  # type: int

    #   rand bit   [31:0] addr
    #   rand logic [31:0] data

    
    async def body(self):
        #reg_block_slave model

        #sv.cast(model, self.model)
        model = self.model

        for i in range(1):
            default_map = model.INDEX.get_default_map()
            status = []
            idx = sv.urandom_range(0, 255)
            await model.INDEX.write(status, idx, _map=default_map, parent=self)

            status = []
            await model.INDEX.mirror(status, UVM_CHECK, UVM_FRONTDOOR,
                    default_map, self)
            got = model.INDEX.get()
            if idx != got:
                uvm_error("IDX_ERR", "exp: " + str(idx) + ' got: ' + str(got))

            for i in range(5):
                status = []
                idx = sv.urandom_range(0, (1 << 64) - 1)
                await model.SESSION[i].SRC.write(status, idx, _map=default_map, parent=self)
                status = []
                await model.SESSION[i].SRC.mirror(status, UVM_CHECK, UVM_FRONTDOOR,
                        default_map, self)

        # Randomize the content of 10 random indexed registers
        #for i in range(10):
        for i in range(1):
            idx = sv.urandom_range(0, self.model.nsession-1)
            data = sv.urandom()
            status = []
            default_map = model.TABLES[idx].get_default_map()
            print("Writing to TABLES idx " + str(idx))
            await model.TABLES[idx].write(status, data, _map=default_map, parent=self)
            print("AFTER Writing to TABLES idx " + str(idx))

        # Find which indexed registers are non-zero
        #for i in range(len(model.TABLES)):
        for i in range(1):
            data = []
            status = []

            print("Reading from TABLES idx " + str(i))
            await model.TABLES[i].read(status, data)
            if data[0] != 0:
                print(sv.sformatf("TABLES[%0d] is 0x%h...", i, data[0]))


uvm_object_utils(user_test_seq)



#
#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------

# File list only for reference at the moment:

#`include "tlm1/uvm_tlm_ifs.svh"
#`include "tlm1/uvm_sqr_ifs.svh"
#`include "base/uvm_port_base.svh"
#
#`include "tlm1/uvm_tlm_imps.svh"
#
#`include "tlm1/uvm_imps.svh"
#`include "tlm1/uvm_ports.svh"
#`include "tlm1/uvm_exports.svh"
#`include "tlm1/uvm_analysis_port.svh"
#
#`include "tlm1/uvm_tlm_fifo_base.svh"
#`include "tlm1/uvm_tlm_fifos.svh"
#`include "tlm1/uvm_tlm_req_rsp.svh"
#
#`include "tlm1/uvm_sqr_connections.svh"
#



#
#----------------------------------------------------------------------
#   Copyright 2019      Tuomas Poikela
#
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#
#----------------------------------------------------------------------

from .uvm_queue import UVMQueue
from cocotb.triggers import Event, Timer
from .uvm_debug import UVMDebug, uvm_debug
from typing import List, Any


def _uvm_debug(self, func, msg):
    if self.debug_enabled or UVMDebug.DEBUG:
        uvm_debug(self, func, msg)

OutputItem = List[Any]

class UVMMailbox():
    """
    Class to mimic SystemVerilog mailbox class.
    """

    def __init__(self, size=0, name='mailbox'):
        self.max_size = size
        self.name = name
        self.m_queue = UVMQueue()
        self.m_read_event = Event('mailbox_read_event')
        self.m_write_event = Event('mailbox_write_event')
        self.debug_enabled = False


    async def put(self, item: Any):
        _uvm_debug(self, 'put', 'Starting to check can_put()')
        can_put = self.can_put()
        while can_put is False:
            self.m_read_event.clear()
            await self.m_read_event.wait()
            self.m_read_event.clear()
            can_put = self.can_put()
        _uvm_debug(self, 'put', 'Putting an event into item queue')
        self.m_queue.push_back(item)
        _uvm_debug(self, 'put', 'ZZZ pushed to queue, can_get is ' + str(self.can_get()))
        self.m_write_event.set()
        _uvm_debug(self, 'put', 'Finished')


    async def get(self, itemq: OutputItem):
        can_get = self.can_get()
        while can_get is False:
            _uvm_debug(self, 'get', 'waiting write event to get item')
            self.m_write_event.clear()
            await self.m_write_event.wait()
            self.m_write_event.clear()
            _uvm_debug(self, 'get', 'event cleared, can_get ' + str(self.can_get()))
            can_get = self.can_get()

        _uvm_debug(self, 'get', 'wait write event DONE')
        if not self.can_get():
            raise Exception('can_get() should return True, but got false. q: ' +
                    str(self.m_queue))
        item = self.m_queue.pop_front()
        self.m_read_event.set()
        _uvm_debug(self, 'get', 'getting an item from mailbox now')
        itemq.append(item)


    async def peek(self, itemq: OutputItem):
        """ Peeks (with blocking) next item from mailbox without removing it """
        if not self.can_get():
            _uvm_debug(self, 'get', 'waiting write event to get item')
            self.m_write_event.clear()
            await self.m_write_event.wait()
            self.m_write_event.clear()
            _uvm_debug(self, 'get', 'event cleared, can_get ' + str(self.can_get()))
        item = self.m_queue.front()
        itemq.append(item)

    def try_put(self, item: Any) -> bool:
        _uvm_debug(self, 'try_put', 'Starting function')
        if self.can_put() is True:
            _uvm_debug(self, 'try_put', 'can_put is True')
            self.m_queue.push_back(item)
            _uvm_debug(self, 'try_put', 'pushed to queue, can_get is ' + str(self.can_get()))
            self.m_write_event.set()
            _uvm_debug(self, 'try_put', 'try_put finishing OK')
            return True
        return False

    def try_get(self, itemq: OutputItem) -> bool:
        """
        Tries to retrieve an item and append it to given list.
        Returns True if success, False otherwise.

        Args:
            itemq: List[Any]
        Returns:
            bool - True if got an item, False otherwise.
        """
        if self.can_get() is True:
            item = self.m_queue.pop_front()
            itemq.append(item)
            self.m_read_event.set()
            _uvm_debug(self, 'try_get', 'try_get finishing OK')
            return True
        return False

    def try_peek(self, itemq: OutputItem) -> bool:
        """
        Tries to "peek" an item and append it to given list.
        Returns True if success, False otherwise. Does not modify
        contents of the mailbox.

        Args:
            itemq: List[Any]
        Returns:
            bool - True if got an item, False otherwise.
        """
        if self.can_get() is True:
            item = self.m_queue.front()
            itemq.append(item)
            _uvm_debug(self, 'try_peek', 'try_peek finishing OK')
            return True
        return False

    def can_get(self) -> bool:
        return self.m_queue.size() > 0

    def can_put(self) -> bool:
        if self.max_size <= 0:
            return True
        return self.m_queue.size() < self.max_size

    def num(self) -> int:
        return self.m_queue.size() > 0



#
# ------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010-2013 Synopsys, Inc.
#   Copyright 2013      NVIDIA Corporation
#   Copyright 2019      Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
# ------------------------------------------------------------------------------


# This bit marks where filtering should occur to remove uvm stuff from a
# scope
### bit uvm_start_uvm_declarations = 1;

# ------------------------------------------------------------------------------
#
# Section: Types and Enumerations
#
# ------------------------------------------------------------------------------

# ------------------------
# Group: Field automation
# ------------------------


### parameter UVM_STREAMBITS = `UVM_MAX_STREAMBITS;
UVM_STREAMBITS = 256

# Enum: uvm_radix_enum
#
# Specifies the radix to print or record in.
#
# UVM_BIN       - Selects binary (%b) format
# UVM_DEC       - Selects decimal (%d) format
# UVM_UNSIGNED  - Selects unsigned decimal (%u) format
# UVM_UNFORMAT2 - Selects unformatted 2 value data (%u) format
# UVM_UNFORMAT4 - Selects unformatted 4 value data (%z) format
# UVM_OCT       - Selects octal (%o) format
# UVM_HEX       - Selects hexadecimal (%h) format
# UVM_STRING    - Selects string (%s) format
# UVM_TIME      - Selects time (%t) format
# UVM_ENUM      - Selects enumeration value (name) format
# UVM_REAL      - Selects real (%g) in exponential or decimal format,
#                 whichever format results in the shorter printed output
# UVM_REAL_DEC  - Selects real (%f) in decimal format
# UVM_REAL_EXP  - Selects real (%e) in exponential format

UVM_BIN       = 0x1000000
UVM_DEC       = 0x2000000
UVM_UNSIGNED  = 0x3000000
UVM_UNFORMAT2 = 0x4000000
UVM_UNFORMAT4 = 0x5000000
UVM_OCT       = 0x6000000
UVM_HEX       = 0x7000000
UVM_STRING    = 0x8000000
UVM_TIME      = 0x9000000
UVM_ENUM      = 0xa000000
UVM_REAL      = 0xb000000
UVM_REAL_DEC  = 0xc000000
UVM_REAL_EXP  = 0xd000000
UVM_NORADIX   = 0

UVM_RADIX = 0xf000000  # 4 bits setting the radix

UVM_RADIX_TO_STRING_DICT = {
    UVM_BIN: "b",
    UVM_OCT: "o",
    UVM_DEC: "d",
    UVM_HEX: "h",
    UVM_UNSIGNED: "u",
    UVM_UNFORMAT2: "u",
    UVM_UNFORMAT4: "z",
    UVM_STRING: "s",
    UVM_TIME: "t",
    UVM_ENUM: "s",
    UVM_REAL: "g",
    UVM_REAL_DEC: "f",
    UVM_REAL_EXP: "e"
}


def uvm_radix_to_string(radix: int) -> str:
    """
    Function- uvm_radix_to_string

    Returns:
        str: Letter for the given radix constant.
    """
    if radix in UVM_RADIX_TO_STRING_DICT:
        return UVM_RADIX_TO_STRING_DICT[radix]
    return "x"

# Enum: uvm_recursion_policy_enum
#
# Specifies the policy for copying objects.
#
# UVM_DEEP      - Objects are deep copied (object must implement <uvm_object::copy> method)
# UVM_SHALLOW   - Objects are shallow copied using default SV copy.
# UVM_REFERENCE - Only object handles are copied.


UVM_DEFAULT_POLICY = 0
UVM_DEEP           = 0x400
UVM_SHALLOW        = 0x800
UVM_REFERENCE      = 0x1000

# Enum: uvm_active_passive_enum
#
# Convenience value to define whether a component, usually an agent,
# is in "active" mode or "passive" mode.
#
# UVM_PASSIVE - "Passive" mode
# UVM_ACTIVE  - "Active" mode
UVM_PASSIVE = False
UVM_ACTIVE = True

# Parameter: `uvm_field_* macro flags
#
# Defines what operations a given field should be involved in.
# Bitwise OR all that apply.
#
# UVM_DEFAULT   - All field operations turned on
# UVM_COPY      - Field will participate in <uvm_object::copy>
# UVM_COMPARE   - Field will participate in <uvm_object::compare>
# UVM_PRINT     - Field will participate in <uvm_object::print>
# UVM_RECORD    - Field will participate in <uvm_object::record>
# UVM_PACK      - Field will participate in <uvm_object::pack>
#
# UVM_NOCOPY    - Field will not participate in <uvm_object::copy>
# UVM_NOCOMPARE - Field will not participate in <uvm_object::compare>
# UVM_NOPRINT   - Field will not participate in <uvm_object::print>
# UVM_NORECORD  - Field will not participate in <uvm_object::record>
# UVM_NOPACK    - Field will not participate in <uvm_object::pack>
#
# UVM_DEEP      - Object field will be deep copied
# UVM_SHALLOW   - Object field will be shallow copied
# UVM_REFERENCE - Object field will copied by reference
#
# UVM_READONLY  - Object field will NOT be automatically configured.

UVM_MACRO_NUMFLAGS = 17
# A=ABSTRACT Y=PHYSICAL
# F=REFERENCE, S=SHALLOW, D=DEEP
# K=PACK, R=RECORD, P=PRINT, M=COMPARE, C=COPY
# ----------------- AYFSD K R P M C
UVM_DEFAULT     = 0b000010101010101
UVM_ALL_ON      = 0b000000101010101
UVM_FLAGS_ON    = 0b000000101010101
UVM_FLAGS_OFF   = 0

# Values are OR'ed into a 32 bit value
# and externally
UVM_COPY         = (1 << 0)
UVM_NOCOPY       = (1 << 1)
UVM_COMPARE      = (1 << 2)
UVM_NOCOMPARE    = (1 << 3)
UVM_PRINT        = (1 << 4)
UVM_NOPRINT      = (1 << 5)
UVM_RECORD       = (1 << 6)
UVM_NORECORD     = (1 << 7)
UVM_PACK         = (1 << 8)
UVM_NOPACK       = (1 << 9)
# parameter UVM_DEEP         = (1<<10);
# parameter UVM_SHALLOW      = (1<<11);
# parameter UVM_REFERENCE    = (1<<12);
UVM_PHYSICAL     = (1 << 13)
UVM_ABSTRACT     = (1 << 14)
UVM_READONLY     = (1 << 15)
UVM_NODEFPRINT   = (1 << 16)

# Extra values that are used for extra methods
UVM_MACRO_EXTRAS  = (1 << UVM_MACRO_NUMFLAGS)
UVM_FLAGS        = UVM_MACRO_EXTRAS+1
UVM_UNPACK       = UVM_MACRO_EXTRAS+2
UVM_CHECK_FIELDS = UVM_MACRO_EXTRAS+3
UVM_END_DATA_EXTRA = UVM_MACRO_EXTRAS+4

# Get and set methods (in uvm_object). Used by the set/get* functions
# to tell the object what operation to perform on the fields.
UVM_START_FUNCS  = UVM_END_DATA_EXTRA+1
UVM_SET           = UVM_START_FUNCS+1
UVM_SETINT        = UVM_SET
UVM_SETOBJ        = UVM_START_FUNCS+2
UVM_SETSTR        = UVM_START_FUNCS+3
UVM_END_FUNCS     = UVM_SETSTR

# Global string variables
uvm_aa_string_key = ""

# -----------------
# Group: Reporting
# -----------------

# Enum: uvm_severity
#
# Defines all possible values for report severity.
#
#   UVM_INFO    - Informative message.
#   UVM_WARNING - Indicates a potential problem.
#   UVM_ERROR   - Indicates a real problem. Simulation continues subject
#                 to the configured message action.
#   UVM_FATAL   - Indicates a problem from which simulation cannot
#                 recover. Simulation exits via $finish after a #0 delay.

UVM_INFO = 0
UVM_WARNING = 1
UVM_ERROR = 2
UVM_FATAL = 3
UVM_SEVERITY_LEVELS = [UVM_INFO, UVM_WARNING, UVM_ERROR, UVM_FATAL]
UVM_SEVERITY_NAMES = ["UVM_INFO", "UVM_WARNING", "UVM_ERROR", "UVM_FATAL"]

# Enum: uvm_action
#
# Defines all possible values for report actions. Each report is configured
# to execute one or more actions, determined by the bitwise OR of any or all
# of the following enumeration constants.
#
#   UVM_NO_ACTION - No action is taken
#   UVM_DISPLAY   - Sends the report to the standard output
#   UVM_LOG       - Sends the report to the file(s) for this (severity,id) pair
#   UVM_COUNT     - Counts the number of reports with the COUNT attribute.
#                   When this value reaches max_quit_count, the simulation terminates
#   UVM_EXIT      - Terminates the simulation immediately.
#   UVM_CALL_HOOK - Callback the report hook methods
#   UVM_STOP      - Causes ~$stop~ to be executed, putting the simulation into
#                   interactive mode.
#   UVM_RM_RECORD - Sends the report to the recorder

UVM_NO_ACTION = 0b0000000
UVM_DISPLAY   = 0b0000001
UVM_LOG       = 0b0000010
UVM_COUNT     = 0b0000100
UVM_EXIT      = 0b0001000
UVM_CALL_HOOK = 0b0010000
UVM_STOP      = 0b0100000
UVM_RM_RECORD = 0b1000000

# Enum: uvm_verbosity
#
# Defines standard verbosity levels for reports.
#
#  UVM_NONE   - Report is always printed. Verbosity level setting cannot
#               disable it.
#  UVM_LOW    - Report is issued if configured verbosity is set to UVM_LOW
#               or above.
#  UVM_MEDIUM - Report is issued if configured verbosity is set to UVM_MEDIUM
#               or above.
#  UVM_HIGH   - Report is issued if configured verbosity is set to UVM_HIGH
#               or above.
#  UVM_FULL   - Report is issued if configured verbosity is set to UVM_FULL
#               or above.

UVM_NONE   = 0
UVM_LOW    = 100
UVM_MEDIUM = 200
UVM_HIGH   = 300
UVM_FULL   = 400
UVM_DEBUG  = 500

UVM_VERBOSITY_LIST = [UVM_NONE, UVM_LOW, UVM_MEDIUM, UVM_HIGH, UVM_FULL,
        UVM_DEBUG]

### typedef int UVM_FILE;

# -----------------
# Group: Port Type
# -----------------

# Enum: uvm_port_type_e
#
# Specifies the type of port
#
# UVM_PORT           - The port requires the interface that is its type
#                      parameter.
# UVM_EXPORT         - The port provides the interface that is its type
#                      parameter via a connection to some other export or
#                      implementation.
# UVM_IMPLEMENTATION - The port provides the interface that is its type
#                      parameter, and it is bound to the component that
#                      implements the interface.

UVM_PORT = 0
UVM_EXPORT = 1
UVM_IMPLEMENTATION = 2

# -----------------
# Group: Sequences
# -----------------

# Enum: uvm_sequencer_arb_mode
#
# Specifies a sequencer's arbitration mode
#
# UVM_SEQ_ARB_FIFO          - Requests are granted in FIFO order (default)
# UVM_SEQ_ARB_WEIGHTED      - Requests are granted randomly by weight
# UVM_SEQ_ARB_RANDOM        - Requests are granted randomly
# UVM_SEQ_ARB_STRICT_FIFO   - Requests at highest priority granted in fifo order
# UVM_SEQ_ARB_STRICT_RANDOM - Requests at highest priority granted in randomly
# UVM_SEQ_ARB_USER          - Arbitration is delegated to the user-defined
#                             function, user_priority_arbitration. That function
#                             will specify the next sequence to grant.

UVM_SEQ_ARB_FIFO = 0
UVM_SEQ_ARB_WEIGHTED = 1
UVM_SEQ_ARB_RANDOM = 2
UVM_SEQ_ARB_STRICT_FIFO = 3
UVM_SEQ_ARB_STRICT_RANDOM = 4
UVM_SEQ_ARB_USER = 5

### typedef uvm_sequencer_arb_mode UVM_SEQ_ARB_TYPE; // backward compat

# Enum: uvm_sequence_state_enum
#
# Defines current sequence state
#
# UVM_CREATED            - The sequence has been allocated.
# UVM_PRE_START          - The sequence is started and the
#                          <uvm_sequence_base::pre_start()> task is
#                          being executed.
# UVM_PRE_BODY           - The sequence is started and the
#                          <uvm_sequence_base::pre_body()> task is
#                           being executed.
# UVM_BODY               - The sequence is started and the
#                          <uvm_sequence_base::body()> task is
#                          being executed.
# UVM_ENDED              - The sequence has completed the execution of the
#                          <uvm_sequence_base::body()> task.
# UVM_POST_BODY          - The sequence is started and the
#                          <uvm_sequence_base::post_body()> task is
#                           being executed.
# UVM_POST_START         - The sequence is started and the
#                          <uvm_sequence_base::post_start()> task is
#                          being executed.
# UVM_STOPPED            - The sequence has been forcibly ended by issuing a
#                          <uvm_sequence_base::kill()> on the sequence.
# UVM_FINISHED           - The sequence is completely finished executing.

UVM_CREATED   = 1
UVM_PRE_START = 2
UVM_PRE_BODY  = 4
UVM_BODY      = 8
UVM_POST_BODY = 16
UVM_POST_START = 32
UVM_ENDED     = 64
UVM_STOPPED   = 128
UVM_FINISHED  = 256

# Enum: uvm_sequence_lib_mode
#
# Specifies the random selection mode of a sequence library
#
# UVM_SEQ_LIB_RAND  - Random sequence selection
# UVM_SEQ_LIB_RANDC - Random cyclic sequence selection
# UVM_SEQ_LIB_ITEM  - Emit only items, no sequence execution
# UVM_SEQ_LIB_USER  - Apply a user-defined random-selection algorithm

UVM_SEQ_LIB_RAND = 0
UVM_SEQ_LIB_RANDC = 1
UVM_SEQ_LIB_ITEM = 2
UVM_SEQ_LIB_USER = 3

# ---------------
# Group: Phasing
# ---------------

# Enum: uvm_phase_type
#
# This is an attribute of a <uvm_phase> object which defines the phase
# type.
#
#   UVM_PHASE_IMP      - The phase object is used to traverse the component
#                        hierarchy and call the component phase method as
#                        well as the ~phase_started~ and ~phase_ended~ callbacks.
#                        These nodes are created by the phase macros,
#                        `uvm_builtin_task_phase, `uvm_builtin_topdown_phase,
#                        and `uvm_builtin_bottomup_phase. These nodes represent
#                        the phase type, i.e. uvm_run_phase, uvm_main_phase.
#
#   UVM_PHASE_NODE     - The object represents a simple node instance in
#                        the graph. These nodes will contain a reference to
#                        their corresponding IMP object.
#
#   UVM_PHASE_SCHEDULE - The object represents a portion of the phasing graph,
#                        typically consisting of several NODE types, in series,
#                        parallel, or both.
#
#   UVM_PHASE_TERMINAL - This internal object serves as the termination NODE
#                        for a SCHEDULE phase object.
#
#   UVM_PHASE_DOMAIN   - This object represents an entire graph segment that
#                        executes in parallel with the 'run' phase.
#                        Domains may define any network of NODEs and
#                        SCHEDULEs. The built-in domain, ~uvm~, consists
#                        of a single schedule of all the run-time phases,
#                        starting with ~pre_reset~ and ending with
#                        ~post_shutdown~.
#
UVM_PHASE_IMP = 0
UVM_PHASE_NODE = 1
UVM_PHASE_TERMINAL = 2
UVM_PHASE_SCHEDULE = 3
UVM_PHASE_DOMAIN = 4
UVM_PHASE_GLOBAL = 5


# Enum: uvm_phase_state
# ---------------------
#
# The set of possible states of a phase. This is an attribute of a schedule
# node in the graph, not of a phase, to maintain independent per-domain state
#
#   UVM_PHASE_UNINITIALIZED - The state is uninitialized.  This is the default
#             state for phases, and for nodes which have not yet been added to
#             a schedule.
#
#   UVM_PHASE_DORMANT -  The schedule is not currently operating on the phase
#             node, however it will be scheduled at some point in the future.
#
#   UVM_PHASE_SCHEDULED - At least one immediate predecessor has completed.
#              Scheduled phases block until all predecessors complete or
#              until a jump is executed.
#
#   UVM_PHASE_SYNCING - All predecessors complete, checking that all synced
#              phases (e.g. across domains) are at or beyond this point
#
#   UVM_PHASE_STARTED - phase ready to execute, running phase_started() callback
#
#   UVM_PHASE_EXECUTING - An executing phase is one where the phase callbacks are
#              being executed. Its process is tracked by the phaser.
#
#   UVM_PHASE_READY_TO_END - no objections remain in this phase or in any
#              predecessors of its successors or in any sync'd phases. This
#              state indicates an opportunity for any phase that needs extra
#              time for a clean exit to raise an objection, thereby causing a
#              return to UVM_PHASE_EXECUTING.  If no objection is raised, state
#              will transition to UVM_PHASE_ENDED after a delta cycle.
#              (An example of predecessors of successors: The successor to
#              phase 'run' is 'extract', whose predecessors are 'run' and
#              'post_shutdown'. Therefore, 'run' will go to this state when
#              both its objections and those of 'post_shutdown' are all dropped.
#
#   UVM_PHASE_ENDED - phase completed execution, now running phase_ended() callback
#
#   UVM_PHASE_JUMPING - all processes related to phase are being killed and all
#                       predecessors are forced into the DONE state.
#
#   UVM_PHASE_CLEANUP - all processes related to phase are being killed
#
#   UVM_PHASE_DONE - A phase is done after it terminated execution.  Becoming
#              done may enable a waiting successor phase to execute.
#
#    The state transitions occur as follows:
#
# |   UNINITIALIZED -> DORMANT -> SCHED -> SYNC -> START -> EXEC -> READY -> END -+-> CLEAN -> DONE
# |                       ^                                                       |
# |                       |                      <-- jump_to                      |
# |                       +-------------------------------------------- JUMPING< -+

UVM_PHASE_UNINITIALIZED = 0
UVM_PHASE_DORMANT      = 1
UVM_PHASE_SCHEDULED    = 2
UVM_PHASE_SYNCING      = 4
UVM_PHASE_STARTED      = 8
UVM_PHASE_EXECUTING    = 16
UVM_PHASE_READY_TO_END = 32
UVM_PHASE_ENDED        = 64
UVM_PHASE_CLEANUP      = 128
UVM_PHASE_DONE         = 256
UVM_PHASE_JUMPING      = 512

UVM_PHASE2STR = {
    0: 'UVM_PHASE_UNINITIALIZED',
    1: 'UVM_PHASE_DORMANT ',
    2: 'UVM_PHASE_SCHEDULED ',
    4: 'UVM_PHASE_SYNCING ',
    8: 'UVM_PHASE_STARTED ',
    16: 'UVM_PHASE_EXECUTING ',
    32: 'UVM_PHASE_READY_TO_END',
    64: 'UVM_PHASE_ENDED ',
    128: 'UVM_PHASE_CLEANUP ',
    256: 'UVM_PHASE_DONE ',
    512: 'UVM_PHASE_JUMPING '
}

# Enum: uvm_wait_op
#
# Specifies the operand when using methods like <uvm_phase::wait_for_state>.
#
# UVM_EQ  - equal
# UVM_NE  - not equal
# UVM_LT  - less than
# UVM_LTE - less than or equal to
# UVM_GT  - greater than
# UVM_GTE - greater than or equal to
#
UVM_LT = 0
UVM_LTE = 1
UVM_NE = 2
UVM_EQ = 3
UVM_GT = 4
UVM_GTE = 5

# ------------------
# Group: Objections
# ------------------

# Enum: uvm_objection_event
#
# Enumerated the possible objection events one could wait on. See
# <uvm_objection::wait_for>.
#
# UVM_RAISED      - an objection was raised
# UVM_DROPPED     - an objection was raised
# UVM_ALL_DROPPED - all objections have been dropped
#
UVM_RAISED = 1
UVM_DROPPED = 2
UVM_ALL_DROPPED = 4

# Variable: uvm_default_table_printer
#
# The table printer is a global object that can be used with
# <uvm_object::do_print> to get tabular style printing.

#uvm_default_table_printer = UVMTablePrinter()

# Variable: uvm_default_tree_printer
#
# The tree printer is a global object that can be used with
# <uvm_object::do_print> to get multi-line tree style printing.

### uvm_tree_printer uvm_default_tree_printer  = new();


# Variable: uvm_default_line_printer
#
# The line printer is a global object that can be used with
# <uvm_object::do_print> to get single-line style printing.

### uvm_line_printer uvm_default_line_printer  = new();


# Variable: uvm_default_printer
#
# The default printer policy. Used when calls to <uvm_object::print>
# or <uvm_object::sprint> do not specify a printer policy.
#
# The default printer may be set to any legal <uvm_printer> derived type,
# including the global line, tree, and table printers described above.

### uvm_printer uvm_default_printer = uvm_default_table_printer;


# Variable: uvm_default_packer
#
# The default packer policy. Used when calls to <uvm_object::pack>
# and <uvm_object::unpack> do not specify a packer policy.

### uvm_packer uvm_default_packer = new();


# Variable: uvm_default_comparer
#
#
# The default compare policy. Used when calls to <uvm_object::compare>
# do not specify a comparer policy.

### uvm_comparer uvm_default_comparer = new(); // uvm_comparer::init();



#
#------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------

from ..base.uvm_object_globals import UVM_ACTIVE
from ..base.uvm_component import UVMComponent
from ..base.uvm_resource import UVMResourcePool
from ..base.uvm_queue import UVMQueue



class UVMAgent(UVMComponent):
    """
    The `UVMAgent` virtual class should be used as the base class for the user-
    defined agents. Deriving from `UVMAgent` will allow you to distinguish agents
    from other component types also using its inheritance. Such agents will
    automatically inherit features that may be added to `UVMAgent` in the future.
    
    While an agent's build function, inherited from `UVMComponent`, can be
    implemented to define any agent topology, an agent typically contains three
    subcomponents: a driver, sequencer, and monitor. If the agent is active,
    subtypes should contain all three subcomponents. If the agent is passive,
    subtypes should contain only the monitor.
    """

    def __init__(self, name, parent):
        """
        Creates and initializes an instance of this class using the normal
        constructor arguments for `UVMComponent`: ~name~ is the name of the
        instance, and ~parent~ is the handle to the hierarchical parent, if any.
        
        The int configuration parameter `is_active` is used to identify whether this
        agent should be acting in active or passive mode. This parameter can
        be set by doing::

            UVMConfigDb.set(self, "<relative_path_to_agent>, "is_active", UVM_ACTIVE)
            # Or inside top-env:
            my_agent.is_active = UVM_ACTIVE
        """
        UVMComponent.__init__(self, name, parent)
        self.is_active = UVM_ACTIVE


    def build_phase(self, phase):
        self.active = 0
        rp = None  # uvm_resource_pool
        rq = UVMQueue()  # uvm_resource_types::rsrc_q_t

        UVMComponent.build_phase(self, phase)
        # is_active is treated as if it were declared via `uvm_field_enum,
        # which means it matches against uvm_active_passive_enum, int,
        # int unsigned, uvm_integral_t, uvm_bitstream_t, and string.
        rp = UVMResourcePool.get()
        rq = rp.lookup_name(self.get_full_name(), "is_active", None, 0)
        rq = UVMResourcePool.sort_by_precedence(rq)
        for i in range(len(rq)):
            rsrc = rq[i]  # uvm_resource_base
            rap = rsrc
            self.is_active = rap.read(self)

    type_name = "uvm_agent"

    def get_type_name(self):
        return UVMAgent.type_name

    def get_is_active(self):
        """
        Returns UVM_ACTIVE is the agent is acting as an active agent and
        UVM_PASSIVE if it is acting as a passive agent. The default implementation
        is to just return the is_active flag, but the component developer may
        override this behavior if a more complex algorithm is needed to determine
        the active/passive nature of the agent.
        """
        return self.is_active



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010-2011 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import RisingEdge

from uvm.comps.uvm_test import UVMTest
from uvm import (UVMCoreService, sv, uvm_top, UVMCmdlineProcessor, UVMSequence,
        UVMConfigDb, UVMResourceDb, UVMReportServer)
from uvm.reg import (uvm_reg_sequence)
from uvm.macros import *
from tb_env import tb_env
from user_test import user_test_seq


class dut_reset_seq(UVMSequence):

    def __init__(self, name="dut_reset_seq"):
        super().__init__(name)
        self.vif = None

    
    async def body(self):
        self.vif.rst <= 1
        for i in range(5):
            await RisingEdge(self.vif.clk)
        self.vif.rst <= 0


uvm_object_utils(dut_reset_seq)


class cmdline_test(UVMTest):
    #
    #
    def __init__(self, name, parent):
        super().__init__(name, parent)


    
    async def run_phase(self, phase):
        cs_ = UVMCoreService.get()
        env = None  # tb_env
        phase.raise_objection(self)
        arr = []
        if sv.cast(arr, uvm_top.find("env$"), tb_env):
            env = arr[0]

        arr = []
        vif = None
        if UVMConfigDb.get(env, "apb", "vif", arr):
            vif = arr[0]
        else:
            uvm_fatal("NO_VIF", "Could not find apb_vif from config_db")
        # dut_reset_seq rst_seq
        rst_seq = dut_reset_seq.type_id.create("rst_seq", self)
        rst_seq.vif = vif
        await rst_seq.start(None)
        env.model.reset()

        # uvm_cmdline_processor
        opts = UVMCmdlineProcessor.get_inst()
        seq = None  # uvm_reg_sequence
        factory = cs_.get_factory()
        seq_name = []
        opts.get_arg_value("+UVM_REG_SEQ=", seq_name)
        seq_name = seq_name[0]

        seq = factory.create_object_by_name(seq_name, self.get_full_name(), "seq")
        #print("Created obj is " + str(created_obj))
        #if not sv.cast(seq, created_obj, uvm_reg_sequence) or seq is None:
        #    uvm_fatal("TEST/CMD/BADSEQ", "Sequence " + seq_name + " is not a known sequence " + str(seq))
        seq.model = env.model

        UVMResourceDb.set("REG::" + env.model.DMA_RAM.get_full_name() + ".*", "NO_REG_TESTS", 1, self)
        for i in range(len(env.model.SESSION)):
            UVMResourceDb.set("REG::" + env.model.SESSION[i].get_full_name() + ".*", "NO_REG_TESTS", 1, self)
        for i in range(len(env.model.TABLES)):
            UVMResourceDb.set("REG::" + env.model.TABLES[i].get_full_name() + ".*", "NO_REG_TESTS", 1, self)
        await seq.start(None)
        phase.drop_objection(self)

    def check_phase(self, phase):
        super().check_phase(phase)
        server = UVMCoreService.get().get_report_server()
        errors = server.get_severity_count(UVM_ERROR)
        if errors > 0:
            uvm_fatal("FOUND ERRORS", "There were " + str(errors)
            + " UVM_ERRORs in the test")


uvm_component_utils(cmdline_test)



#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------

from typing import Any

from ..base.uvm_transaction import UVMTransaction
from ..base.uvm_globals import *


UVMSequencerBase = Any
UVMSequenceBase = Any

#------------------------------------------------------------------------------
#
# CLASS: UVMSequenceItem
#
# The base class for user-defined sequence items and also the base class for
# the uvm_sequence class. The UVMSequenceItem class provides the basic
# functionality for objects, both sequence items and sequences, to operate in
# the sequence mechanism.
#
#------------------------------------------------------------------------------

class UVMSequenceItem(UVMTransaction):
    #  protected  uvm_sequencer_base self.m_sequencer
    #  protected  uvm_sequence_base  self.m_parent_sequence
    #  bit        self.print_sequence_info

    #  static     bit issued1,issued2
    issued1 = False
    issued2 = False

    def __init__(self, name="UVMSequenceItem") -> None:
        """
        Function: new

        The constructor method for UVMSequenceItem.

        Args:
            name:
        """
        UVMTransaction.__init__(self, name)
        self.m_sequence_id = -1
        self.m_use_sequence_info = False
        self.m_depth = -1
        self.m_sequencer: UVMSequencerBase = None
        self.p_sequencer: UVMSequenceBase = None
        self.m_parent_sequence = None
        self.print_sequence_info = False
    #  endfunction

    def get_type_name(self) -> str:
        return "UVMSequenceItem"

    def set_sequence_id(self, id: int) -> None:
        """
          Function- set_sequence_id
        Args:
            id:
        """
        self.m_sequence_id = id

    def get_sequence_id(self) -> int:
        """
          Function: get_sequence_id

          private

          Get_sequence_id is an internal method that is not intended for user code.
          The sequence_id is not a simple integer.  The get_transaction_id is meant
          for users to identify specific transactions.

          These methods allow access to the sequence_item sequence and transaction
          IDs. get_transaction_id and set_transaction_id are methods on the
          uvm_transaction base_class. These IDs are used to identify sequences to
          the sequencer, to route responses back to the sequence that issued a
          request, and to uniquely identify transactions.

          The sequence_id is assigned automatically by a sequencer when a sequence
          initiates communication through any sequencer calls (i.e. `uvm_do_*,
          wait_for_grant).  A sequence_id will remain unique for this sequence
          until it ends or it is killed.  However, a single sequence may have
          multiple valid sequence ids at any point in time.  Should a sequence
          start again after it has ended, it will be given a new unique sequence_id.

          The transaction_id is assigned automatically by the sequence each time a
          transaction is sent to the sequencer with the transaction_id in its
          default (-1) value.  If the user sets the transaction_id to any non-default
          value, that value will be maintained.

          Responses are routed back to this sequences based on sequence_id. The
          sequence may use the transaction_id to correlate responses with their
          requests.
        Returns:
        """
        return (self.m_sequence_id)


    def set_item_context(self, parent_seq, sequencer=None) -> None:
        """
        Function: set_item_context

        Set the sequence and sequencer execution context for a sequence item

        Args:
            parent_seq:
            sequencer:
        """
        self.set_use_sequence_info(1)
        if parent_seq is not None:
            self.set_parent_sequence(parent_seq)
        if (sequencer is None and self.m_parent_sequence is not None):
            sequencer = self.m_parent_sequence.get_sequencer()
        self.set_sequencer(sequencer)
        if (self.m_parent_sequence is not None):
            self.set_depth(self.m_parent_sequence.get_depth() + 1)
        self.reseed()

    def set_use_sequence_info(self, value) -> None:
        """
        Function: set_use_sequence_info

        Args:
            value:
        """
        self.m_use_sequence_info = value


    def get_use_sequence_info(self) -> bool:
        """
        Function: get_use_sequence_info

        These methods are used to set and get the status of the use_sequence_info
        bit. Use_sequence_info controls whether the sequence information
        (sequencer, parent_sequence, sequence_id, etc.) is printed, copied, or
        recorded. When use_sequence_info is the default value of 0, then the
        sequence information is not used. When use_sequence_info is set to 1,
        the sequence information will be used in printing and copying.
        """
        return (self.m_use_sequence_info)


    def set_id_info(self, item: 'UVMSequenceItem') -> None:
        """
        Function: set_id_info

        Copies the sequence_id and transaction_id from the referenced item into
        the calling item.  This routine should always be used by drivers to
        initialize responses for future compatibility.

        Args:
            item:
        """
        if item is None:
            uvm_report_fatal(self.get_full_name(),
                "set_id_info called with None parameter", UVM_NONE)
        self.set_transaction_id(item.get_transaction_id())
        self.set_sequence_id(item.get_sequence_id())

    #  // Function: set_sequencer
    #  //
    #  // Sets the default sequencer for the sequence to sequencer.  It will take
    #  // effect immediately, so it should not be called while the sequence is
    #  // actively communicating with the sequencer.

    def set_sequencer(self, sequencer: UVMSequencerBase) -> None:
        self.m_sequencer = sequencer
        self.m_set_p_sequencer()


    #  // Function: get_sequencer
    #  //
    #  // Returns a reference to the default sequencer used by this sequence.

    def get_sequencer(self) -> UVMSequencerBase:
        return self.m_sequencer


    #  // Function: set_parent_sequence
    #  //
    #  // Sets the parent sequence of this sequence_item.  This is used to identify
    #  // the source sequence of a sequence_item.

    def set_parent_sequence(self, parent: UVMSequenceBase) -> None:
        self.m_parent_sequence = parent


    def get_parent_sequence(self) -> UVMSequenceBase:
        """
        Function: get_parent_sequence

        Returns a reference to the parent sequence of any sequence on which this
        method was called. If this is a parent sequence, the method returns `None`.

        Returns:
            UVMSequenceBase: Parent sequence of this item.
        """
        return (self.m_parent_sequence)


    def set_depth(self, value: int) -> None:
        """
          Function: set_depth

          The depth of any sequence is calculated automatically.  However, the user
          may use  set_depth to specify the depth of a particular sequence. This
          method will override the automatically calculated depth, even if it is
          incorrect.

        Args:
            value:
        """
        self.m_depth = value
    #  endfunction

    def get_depth(self) -> int:
        """
          Function: get_depth

          Returns the depth of a sequence from its parent.  A  parent sequence will
          have a depth of 1, its child will have a depth  of 2, and its grandchild
          will have a depth of 3.

        Returns:
        """

        # If depth has been set or calculated, then use that
        if self.m_depth != -1:
            return (self.m_depth)

        # Calculate the depth, store it, and return the value
        if self.m_parent_sequence is None:
            self.m_depth = 1
        else:
            self.m_depth = self.m_parent_sequence.get_depth() + 1

        return self.m_depth


    def is_item(self) -> bool:
        """
        Function: is_item

        This function may be called on any sequence_item or sequence. It will
        return 1 for items and 0 for sequences (which derive from this class).

        Returns:
        """
        return True


    def get_full_name(self) -> str:
        """
        Function- get_full_name

        Internal method; overrides must follow same naming convention

        Returns:
        """
        get_full_name = ""
        if self.m_parent_sequence is not None:
            get_full_name = self.m_parent_sequence.get_full_name() + "."
        elif self.m_sequencer is not None:
            get_full_name = self.m_sequencer.get_full_name() + "."
        if self.get_name() != "":
            get_full_name = get_full_name + self.get_name()
        else:
            get_full_name = get_full_name + "_item"
        return get_full_name


    def get_root_sequence_name(self) -> str:
        """
          Function: get_root_sequence_name

          Provides the name of the root sequence (the top-most parent sequence).

        Returns:
        """
        #uvm_sequence_base root_seq
        root_seq = self.get_root_sequence()
        if root_seq is None:
            return ""
        else:
            return root_seq.get_name()
    #  endfunction

    def m_set_p_sequencer(self) -> None:
        """
          Function- m_set_p_sequencer

          Internal method
        """
        self.p_sequencer = self.m_sequencer

    def get_root_sequence(self) -> Any:
        """
          Function: get_root_sequence

          Provides a reference to the root sequence (the top-most parent sequence).

        Returns:
        """
        root_seq_base = None  # UVMSequenceItem
        root_seq = None
        root_seq_base = self
        while True:
            if (root_seq_base.get_parent_sequence() is not None):
                root_seq_base = root_seq_base.get_parent_sequence()
                #$cast(root_seq, root_seq_base)
                root_seq = root_seq_base
            else:
                return root_seq


    def get_sequence_path(self) -> str:
        """
        Function: get_sequence_path

        Provides a string of names of each sequence in the full hierarchical
        path. A "." is used as the separator between each sequence.

        Returns:
        """
        this_item = None  # UVMSequenceItem
        seq_path = ""
        this_item = self
        seq_path = self.get_name()
        while True:
            if this_item.get_parent_sequence() is not None:
                this_item = this_item.get_parent_sequence()
                seq_path = this_item.get_name() + "." + seq_path
            else:
                return seq_path


    #  //---------------------------
    #  // Group: Reporting Interface
    #  //---------------------------
    #  //
    #  // Sequence items and sequences will use the sequencer which they are
    #  // associated with for reporting messages. If no sequencer has been se
    #  // for the item/sequence using <set_sequencer> or indirectly via
    #  // <uvm_sequence_base::start_item> or <uvm_sequence_base::start>),
    #  // then the global reporter will be used.
    #
    #  virtual function uvm_report_object uvm_get_report_object()
    #    if(self.m_sequencer == None):
    #      uvm_coreservice_t cs = uvm_coreservice_t::get()
    #       return cs.get_root()
    #    end else
    #      return self.m_sequencer
    #  endfunction
    #
    #  function int uvm_report_enabled(int verbosity,
    #    				  uvm_severity severity=UVM_INFO, string id="")
    #    uvm_report_object l_report_object = uvm_get_report_object()
    #    if (l_report_object.get_report_verbosity_level(severity, id) < verbosity)
    #      return 0
    #    return 1
    #  endfunction
    #
    #  // Function: uvm_report
    #  virtual function void uvm_report( uvm_severity severity,
    #                                    string id,
    #                                    string message,
    #                                    int verbosity = (severity == uvm_severity'(UVM_ERROR)) ? UVM_LOW :
    #                                                    (severity == uvm_severity'(UVM_FATAL)) ? UVM_NONE : UVM_MEDIUM,
    #                                    string filename = "",
    #                                    int line = 0,
    #                                    string context_name = "",
    #                                    bit report_enabled_checked = 0)
    #    uvm_report_message l_report_message
    #    if (report_enabled_checked == 0):
    #      if (!uvm_report_enabled(verbosity, severity, id))
    #        return
    #    end
    #    l_report_message = uvm_report_message::new_report_message()
    #    l_report_message.set_report_message(severity, id, message,
    #					verbosity, filename, line, context_name)
    #    uvm_process_report_message(l_report_message)
    #
    #  endfunction
    #
    #  // Function: uvm_report_info
    #
    #  virtual function void uvm_report_info( string id,
    #					 string message,
    #   					 int verbosity = UVM_MEDIUM,
    #					 string filename = "",
    #					 int line = 0,
    #   					 string context_name = "",
    #					 bit report_enabled_checked = 0)
    #
    #    this.uvm_report(UVM_INFO, id, message, verbosity, filename, line,
    #                    context_name, report_enabled_checked)
    #  endfunction
    #
    #  // Function: uvm_report_warning
    #
    #  virtual function void uvm_report_warning( string id,
    #					    string message,
    #   					    int verbosity = UVM_MEDIUM,
    #					    string filename = "",
    #					    int line = 0,
    #   					    string context_name = "",
    #					    bit report_enabled_checked = 0)
    #
    #    this.uvm_report(UVM_WARNING, id, message, verbosity, filename, line,
    #                    context_name, report_enabled_checked)
    #  endfunction
    #
    #  // Function: uvm_report_error
    #
    #  virtual function void uvm_report_error( string id,
    #					  string message,
    #   					  int verbosity = UVM_LOW,
    #					  string filename = "",
    #					  int line = 0,
    #   					  string context_name = "",
    #					  bit report_enabled_checked = 0)
    #
    #    this.uvm_report(UVM_ERROR, id, message, verbosity, filename, line,
    #                    context_name, report_enabled_checked)
    #  endfunction
    #
    #  // Function: uvm_report_fatal
    #  //
    #  // These are the primary reporting methods in the UVM. UVMSequenceItem
    #  // derived types delegate these functions to their associated sequencer
    #  // if they have one, or to the global reporter. See <uvm_report_object::Reporting>
    #  // for details on the messaging functions.
    #
    #  virtual function void uvm_report_fatal( string id,
    #					  string message,
    #   					  int verbosity = UVM_NONE,
    #					  string filename = "",
    #					  int line = 0,
    #   					  string context_name = "",
    #					  bit report_enabled_checked = 0)
    #
    #    this.uvm_report(UVM_FATAL, id, message, verbosity, filename, line,
    #                    context_name, report_enabled_checked)
    #  endfunction
    #
    #  virtual function void uvm_process_report_message (uvm_report_message report_message)
    #    uvm_report_object l_report_object = uvm_get_report_object()
    #    report_message.set_report_object(l_report_object)
    #    if (report_message.get_context() == "")
    #      report_message.set_context(get_sequence_path())
    #    l_report_object.m_rh.process_report_message(report_message)
    #  endfunction
    #
    #
    #  // Function- do_print
    #  //
    #  // Internal method
    #
    #  function void do_print (uvm_printer printer)
    #    string temp_str0, temp_str1
    #    int depth = get_depth()
    #    super.do_print(printer)
    #    if(self.print_sequence_info || self.m_use_sequence_info):
    #      printer.print_field_int("depth", depth, $bits(depth), UVM_DEC, ".", "int")
    #      if(self.m_parent_sequence != None):
    #        temp_str0 = self.m_parent_sequence.get_name()
    #        temp_str1 = self.m_parent_sequence.get_full_name()
    #      end
    #      printer.print_string("parent sequence (name)", temp_str0)
    #      printer.print_string("parent sequence (full name)", temp_str1)
    #      temp_str1 = ""
    #      if(self.m_sequencer != None):
    #        temp_str1 = self.m_sequencer.get_full_name()
    #      end
    #      printer.print_string("sequencer", temp_str1)
    #    end
    #  endfunction
    #
    #endclass

# Macro for factory creation
# uvm_object_registry(UVMSequenceItem, "UVMSequenceItem")
